# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for acn protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import Any, Dict, cast

from aea.mail.base_pb2 import DialogueMessage
from aea.mail.base_pb2 import Message as ProtobufMessage
from aea.protocols.base import Message, Serializer

from packages.valory.protocols.acn import acn_pb2
from packages.valory.protocols.acn.custom_types import AgentRecord, StatusBody
from packages.valory.protocols.acn.message import AcnMessage


class AcnSerializer(Serializer):
    """Serialization for the 'acn' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Acn' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(AcnMessage, msg)
        message_pb = ProtobufMessage()
        dialogue_message_pb = DialogueMessage()
        acn_msg = acn_pb2.AcnMessage()

        dialogue_message_pb.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        dialogue_message_pb.dialogue_starter_reference = dialogue_reference[0]
        dialogue_message_pb.dialogue_responder_reference = dialogue_reference[1]
        dialogue_message_pb.target = msg.target

        performative_id = msg.performative
        if performative_id == AcnMessage.Performative.REGISTER:
            performative = acn_pb2.AcnMessage.Register_Performative()  # type: ignore
            record = msg.record
            AgentRecord.encode(performative.record, record)
            acn_msg.register.CopyFrom(performative)
        elif performative_id == AcnMessage.Performative.LOOKUP_REQUEST:
            performative = acn_pb2.AcnMessage.Lookup_Request_Performative()  # type: ignore
            agent_address = msg.agent_address
            performative.agent_address = agent_address
            acn_msg.lookup_request.CopyFrom(performative)
        elif performative_id == AcnMessage.Performative.LOOKUP_RESPONSE:
            performative = acn_pb2.AcnMessage.Lookup_Response_Performative()  # type: ignore
            record = msg.record
            AgentRecord.encode(performative.record, record)
            acn_msg.lookup_response.CopyFrom(performative)
        elif performative_id == AcnMessage.Performative.AEA_ENVELOPE:
            performative = acn_pb2.AcnMessage.Aea_Envelope_Performative()  # type: ignore
            envelope = msg.envelope
            performative.envelope = envelope
            record = msg.record
            AgentRecord.encode(performative.record, record)
            acn_msg.aea_envelope.CopyFrom(performative)
        elif performative_id == AcnMessage.Performative.STATUS:
            performative = acn_pb2.AcnMessage.Status_Performative()  # type: ignore
            body = msg.body
            StatusBody.encode(performative.body, body)
            acn_msg.status.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        dialogue_message_pb.content = acn_msg.SerializeToString()

        message_pb.dialogue_message.CopyFrom(dialogue_message_pb)
        message_bytes = message_pb.SerializeToString()
        return message_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Acn' message.

        :param obj: the bytes object.
        :return: the 'Acn' message.
        """
        message_pb = ProtobufMessage()
        acn_pb = acn_pb2.AcnMessage()
        message_pb.ParseFromString(obj)
        message_id = message_pb.dialogue_message.message_id
        dialogue_reference = (
            message_pb.dialogue_message.dialogue_starter_reference,
            message_pb.dialogue_message.dialogue_responder_reference,
        )
        target = message_pb.dialogue_message.target

        acn_pb.ParseFromString(message_pb.dialogue_message.content)
        performative = acn_pb.WhichOneof("performative")
        performative_id = AcnMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == AcnMessage.Performative.REGISTER:
            pb2_record = acn_pb.register.record
            record = AgentRecord.decode(pb2_record)
            performative_content["record"] = record
        elif performative_id == AcnMessage.Performative.LOOKUP_REQUEST:
            agent_address = acn_pb.lookup_request.agent_address
            performative_content["agent_address"] = agent_address
        elif performative_id == AcnMessage.Performative.LOOKUP_RESPONSE:
            pb2_record = acn_pb.lookup_response.record
            record = AgentRecord.decode(pb2_record)
            performative_content["record"] = record
        elif performative_id == AcnMessage.Performative.AEA_ENVELOPE:
            envelope = acn_pb.aea_envelope.envelope
            performative_content["envelope"] = envelope
            pb2_record = acn_pb.aea_envelope.record
            record = AgentRecord.decode(pb2_record)
            performative_content["record"] = record
        elif performative_id == AcnMessage.Performative.STATUS:
            pb2_body = acn_pb.status.body
            body = StatusBody.decode(pb2_body)
            performative_content["body"] = body
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return AcnMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/liquidity_rebalancing_abci skill behaviours with Hardhat."""

import json
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from aea.helpers.transaction.base import RawTransaction, State
from aea.skills.base import Handler
from web3 import Web3

from packages.valory.contracts.gnosis_safe.contract import SafeOperation
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.liquidity_rebalancing_abci.behaviours import (
    EnterPoolTransactionHashBehaviour,
    ExitPoolTransactionHashBehaviour,
    LiquidityRebalancingConsensusBehaviour,
    SAFE_TX_GAS_ENTER,
    SAFE_TX_GAS_EXIT,
    SAFE_TX_GAS_SWAP_BACK,
    SwapBackTransactionHashBehaviour,
    parse_tx_token_balance,
)
from packages.valory.skills.liquidity_rebalancing_abci.handlers import (
    ContractApiHandler,
    HttpHandler,
    LedgerApiHandler,
    SigningHandler,
)
from packages.valory.skills.liquidity_rebalancing_abci.rounds import (
    SynchronizedData as LiquidityRebalancingSynchronizedSata,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    hash_payload_to_hex,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TransactionSettlementSynchronizedSata,
)

from tests.conftest import ROOT_DIR
from tests.test_skills.base import FSMBehaviourBaseCase
from tests.test_skills.integration import (
    AMMIntegrationBaseCase,
    ExpectedContentType,
    ExpectedTypesType,
)
from tests.test_skills.test_liquidity_rebalancing_abci.test_behaviours import (
    A_WETH_POOL_ADDRESS,
    B_WETH_POOL_ADDRESS,
    DEFAULT_MINTER,
    LP_TOKEN_ADDRESS,
    TOKEN_A_ADDRESS,
    TOKEN_B_ADDRESS,
    WETH_ADDRESS,
    get_default_strategy,
)


class LiquidityRebalancingBehaviourBaseCase(FSMBehaviourBaseCase):
    """Base case for testing LiquidityRebalancing FSMBehaviour."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "valory", "skills", "liquidity_provision_abci"
    )

    behaviour: LiquidityRebalancingConsensusBehaviour
    ledger_handler: LedgerApiHandler
    http_handler: HttpHandler
    contract_handler: ContractApiHandler
    signing_handler: SigningHandler


class LiquidityProvisionIntegrationBaseCase(
    LiquidityRebalancingBehaviourBaseCase, AMMIntegrationBaseCase
):
    """Base case for integration testing `LiquidityProvision` FSM Behaviour."""

    strategy: Dict
    multisend_contract_address: str = "0x2279B7A0a67DB372996a5FaB50D91eAA73d2eBe6"
    router_contract_address: str = "0xA51c1fc2f0D1a1b8494Ed1FE312d7C3a78Ed91C0"
    enter_nonce: int
    exit_nonce: int
    swap_back_nonce: int
    default_synchronized_data_hash: LiquidityRebalancingSynchronizedSata

    @classmethod
    def setup(cls, **kwargs: Any) -> None:
        """Setup."""
        super().setup()

        cls.multisend_data_enter = "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"
        cls.multisend_data_exit = "8d80ff0a0000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000026b0050cd56fb094f8f06063066a619d898475dd3eede00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000044095ea7b3000000000000000000000000a51c1fc2f0d1a1b8494ed1fe312d7c3a78ed91c0ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff00a51c1fc2f0d1a1b8494ed1fe312d7c3a78ed91c0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000e4baa2abde0000000000000000000000000dcd1bf9a1b36ce34237eeafef220932846bcd820000000000000000000000009a676e781a523b5d0c0e43731313a708cb60750800000000000000000000000000000000000000000000000000000000000003e80000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000068fcdf52066cce5612827e872c45767e5a1f65510000000000000000000000000000000000000000000000000000000063b0beef0050cd56fb094f8f06063066a619d898475dd3eede00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000044095ea7b3000000000000000000000000a51c1fc2f0d1a1b8494ed1fe312d7c3a78ed91c00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
        cls.multisend_data_swap_back = "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"

        cls.most_voted_tx_hash_enter = (
            "b68291e2144f384f927e423d78702463543dd56e8fc8910b2edd20ddf47bbf46"
        )
        cls.most_voted_tx_hash_exit = (
            "70de1e8c13f56ea56f36fdc9d69635b60f1d0dc7fe4fb4a89024e16aa8ad0e12"
        )
        cls.most_voted_tx_hash_swap_back = (
            "7920b96f55935f49f636143d6071aadc5a7b53670a9ad5ddc16ab1c692a55e1f"
        )

        cls.enter_nonce = 8
        cls.exit_nonce = cls.enter_nonce + 1
        cls.swap_back_nonce = cls.enter_nonce + 2

        # setup default objects
        cls.strategy = get_default_strategy(
            is_base_native=False, is_a_native=False, is_b_native=False
        )
        # corresponds to datetime.datetime(2022, 12, 31, 23, 59, 59) using datetime.datetime.fromtimestamp(.)
        cls.strategy["deadline"] = 1672527599

        cls.default_synchronized_data_hash = LiquidityRebalancingSynchronizedSata(
            AbciAppDB(
                setup_data=AbciAppDB.data_to_lists(
                    dict(
                        safe_contract_address=cls.safe_contract_address,
                        most_voted_keeper_address=cls.keeper_address,
                        most_voted_strategy=json.dumps(cls.strategy),
                        multisend_contract_address=cls.multisend_contract_address,
                        router_contract_address=cls.router_contract_address,
                        participants=frozenset(list(cls.safe_owners.keys())),
                    )
                ),
            )
        )

        keeper_retries = 1
        keepers = next(iter(cls.agents.keys()))
        cls.tx_settlement_synchronized_data = TransactionSettlementSynchronizedSata(
            AbciAppDB(
                setup_data=AbciAppDB.data_to_lists(
                    dict(
                        safe_contract_address=cls.safe_contract_address,
                        most_voted_keeper_address=cls.keeper_address,
                        participants=frozenset(list(cls.safe_owners.keys())),
                        keepers=keeper_retries.to_bytes(32, "big").hex() + keepers,
                    )
                ),
            )
        )

    @staticmethod
    def get_decoded_logs(gnosis_instance: Any, receipt: dict) -> List[Dict]:
        """Get decoded logs."""
        # Find ABI events
        decoded_logs = []
        abi_events = [abi for abi in gnosis_instance.abi if abi["type"] == "event"]
        for logs in receipt["logs"]:
            for receipt_event_signature_hex in logs["topics"]:
                for event in abi_events:
                    # Get event signature components
                    name = event["name"]
                    inputs = [param["type"] for param in event["inputs"]]
                    inputs_ = ",".join(inputs)
                    # Hash event signature
                    event_signature_text = f"{name}({inputs_})"
                    event_signature_hex = Web3.toHex(
                        Web3.keccak(text=event_signature_text)
                    )
                    # Find match between log's event signature and ABI's event signature
                    if event_signature_hex == receipt_event_signature_hex:
                        decoded_log = gnosis_instance.events[
                            event["name"]
                        ]().processReceipt(receipt)
                        decoded_logs.append({name: decoded_log})
        return decoded_logs

    def validate_tx(self, simulate_timeout: bool = False) -> None:
        """Validate the sent transaction."""
        super().validate_tx(simulate_timeout)

        # eventually replace with https://pypi.org/project/eth-event/
        receipt = self.ethereum_api.get_transaction_receipt(
            self.tx_settlement_synchronized_data.to_be_validated_tx_hash
        )
        logs = self.get_decoded_logs(self.gnosis_instance, receipt)

        assert all(
            [key != "ExecutionFailure" for dict_ in logs for key in dict_.keys()]
        )

    def send_and_validate(
        self,
    ) -> None:
        """Send and validate a transaction"""
        # Sign the transaction
        self.sign_tx()
        # Send the transaction
        self.send_tx()
        # Validate the transaction
        self.validate_tx()


class TestLiquidityRebalancingHardhat(LiquidityProvisionIntegrationBaseCase):
    """Test liquidity pool behaviours in a Hardhat environment."""

    def test_full_run(self) -> None:
        """Run the test"""
        timestamp = self.ethereum_api.api.eth.get_block("latest")["timestamp"]
        assert self.strategy["deadline"] > timestamp, "Increase timestamp!"
        strategy = deepcopy(self.strategy)

        # ENTER POOL ------------------------------------------------------

        # Prepare the transaction
        strategy["safe_nonce"] = 0

        synchronized_data_enter_hash = cast(
            LiquidityRebalancingSynchronizedSata,
            self.default_synchronized_data_hash.update(),
        )

        cycles_enter = 8
        handlers_enter: List[Optional[Handler]] = [self.contract_handler] * cycles_enter
        expected_content_enter: ExpectedContentType = [
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            }
        ] * cycles_enter
        expected_types_enter: ExpectedTypesType = [
            {
                "raw_transaction": RawTransaction,
            }
        ] * cycles_enter
        _, _, _, _, _, _, msg_a, msg_b = self.process_n_messages(
            cycles_enter,
            synchronized_data_enter_hash,
            EnterPoolTransactionHashBehaviour.behaviour_id,
            handlers_enter,
            expected_content_enter,
            expected_types_enter,
        )
        assert msg_a is not None and isinstance(msg_a, ContractApiMessage)
        tx_data_enter = cast(str, msg_a.raw_transaction.body["data"])[2:]
        assert tx_data_enter == self.multisend_data_enter
        assert msg_b is not None and isinstance(msg_b, ContractApiMessage)
        tx_hash_enter = cast(str, msg_b.raw_transaction.body["tx_hash"])[2:]
        assert tx_hash_enter == self.most_voted_tx_hash_enter

        payload = hash_payload_to_hex(
            tx_hash_enter,
            0,
            SAFE_TX_GAS_ENTER,
            self.multisend_contract_address,
            bytes.fromhex(self.multisend_data_enter),
            SafeOperation.DELEGATE_CALL.value,
        )

        # update period state with safe's tx hash
        self.tx_settlement_synchronized_data.update(
            most_voted_tx_hash=payload,
        )

        # Sign, send, validate
        self.send_and_validate()

        # EXIT POOL ------------------------------------------------------

        # Prepare the transaction
        strategy["safe_nonce"] = 1

        synchronized_data_exit_hash = cast(
            LiquidityRebalancingSynchronizedSata,
            self.default_synchronized_data_hash.update(
                most_voted_strategy=json.dumps(strategy),
                final_tx_hash=self.tx_settlement_synchronized_data.final_tx_hash,
            ),
        )

        cycles_exit = 6
        handlers_exit: List[Optional[Handler]] = [self.contract_handler] * cycles_exit
        expected_content_exit: ExpectedContentType = [
            {"performative": ContractApiMessage.Performative.STATE},  # type: ignore
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
        ]
        expected_types_exit: ExpectedTypesType = [
            {"state": State},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
        ]
        transfers_msg_enter, _, _, _, msg_a, msg_b = self.process_n_messages(
            cycles_exit,
            synchronized_data_exit_hash,
            ExitPoolTransactionHashBehaviour.behaviour_id,
            handlers_exit,
            expected_content_exit,
            expected_types_exit,
        )
        assert msg_a is not None and isinstance(msg_a, ContractApiMessage)
        tx_data_exit = cast(str, msg_a.raw_transaction.body["data"])[2:]
        assert tx_data_exit == self.multisend_data_exit
        assert msg_b is not None and isinstance(msg_b, ContractApiMessage)
        tx_hash_exit = cast(str, msg_b.raw_transaction.body["tx_hash"])[2:]
        assert tx_hash_exit == self.most_voted_tx_hash_exit

        transfers_enter = cast(ContractApiMessage, transfers_msg_enter).state.body[
            "logs"
        ]

        amount_weth_sent_a = parse_tx_token_balance(
            cast(list, transfers_enter),
            WETH_ADDRESS,
            self.safe_contract_address,
            A_WETH_POOL_ADDRESS,
        )
        assert (
            amount_weth_sent_a == 1004
        ), f"Token base amount sent is not correct (A): {amount_weth_sent_a} != 1004"

        amount_weth_sent_b = parse_tx_token_balance(
            cast(list, transfers_enter),
            WETH_ADDRESS,
            self.safe_contract_address,
            B_WETH_POOL_ADDRESS,
        )
        assert (
            amount_weth_sent_b == 1004
        ), f"Token base amount sent is not correct (B): {amount_weth_sent_b} != 1004"

        amount_a_received = parse_tx_token_balance(
            cast(list, transfers_enter),
            TOKEN_A_ADDRESS,
            A_WETH_POOL_ADDRESS,
            self.safe_contract_address,
        )
        assert (
            amount_a_received == 1000
        ), f"Token A amount received is not correct: {amount_a_received} != 1000"

        amount_b_received = parse_tx_token_balance(
            cast(list, transfers_enter),
            TOKEN_B_ADDRESS,
            B_WETH_POOL_ADDRESS,
            self.safe_contract_address,
        )
        assert (
            amount_b_received == 1000
        ), f"Token B amount received is not correct: {amount_b_received} != 1000"

        amount_a_sent = parse_tx_token_balance(
            cast(list, transfers_enter),
            TOKEN_A_ADDRESS,
            self.safe_contract_address,
            LP_TOKEN_ADDRESS,
        )
        assert (
            amount_a_sent == 1000
        ), f"Token A amount sent is not correct: {amount_a_sent} != 1000"

        amount_b_sent = parse_tx_token_balance(
            cast(list, transfers_enter),
            TOKEN_B_ADDRESS,
            self.safe_contract_address,
            LP_TOKEN_ADDRESS,
        )
        assert (
            amount_b_sent == 1000
        ), f"Token B amount sent is not correct: {amount_b_sent} != 1000"

        amount_lp_received = parse_tx_token_balance(
            cast(list, transfers_enter),
            LP_TOKEN_ADDRESS,
            DEFAULT_MINTER,
            self.safe_contract_address,
        )
        assert (
            amount_lp_received == 1000
        ), f"LP amount received is not correct: {amount_lp_received} != 1000"

        payload = hash_payload_to_hex(
            tx_hash_exit,
            0,
            SAFE_TX_GAS_EXIT,
            self.multisend_contract_address,
            bytes.fromhex(self.multisend_data_exit),
            SafeOperation.DELEGATE_CALL.value,
        )

        # update period state with safe's tx hash
        self.tx_settlement_synchronized_data.update(
            most_voted_tx_hash=payload,
        )

        # Sign, send, validate
        self.send_and_validate()

        # SWAP BACK ------------------------------------------------------

        # Prepare the transaction
        strategy["safe_nonce"] = 2

        synchronized_data_swap_back_hash = cast(
            LiquidityRebalancingSynchronizedSata,
            self.default_synchronized_data_hash.update(
                most_voted_strategy=json.dumps(strategy),
                final_tx_hash=self.tx_settlement_synchronized_data.final_tx_hash,
            ),
        )

        cycles_swap_back = 8
        handlers_swap_back: List[Optional[Handler]] = [
            self.contract_handler
        ] * cycles_swap_back
        expected_content_swap_back: ExpectedContentType = [
            {"performative": ContractApiMessage.Performative.STATE},  # type: ignore
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
            {
                "performative": ContractApiMessage.Performative.RAW_TRANSACTION  # type: ignore
            },
        ]
        expected_types_swap_back: ExpectedTypesType = [
            {"state": State},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
            {"raw_transaction": RawTransaction},
        ]
        transfers_msg_exit, _, _, _, _, _, msg_a, msg_b = self.process_n_messages(
            cycles_swap_back,
            synchronized_data_swap_back_hash,
            SwapBackTransactionHashBehaviour.behaviour_id,
            handlers_swap_back,
            expected_content_swap_back,
            expected_types_swap_back,
        )
        assert msg_a is not None and isinstance(msg_a, ContractApiMessage)
        tx_data_swap_back = cast(str, msg_a.raw_transaction.body["data"])[2:]
        assert tx_data_swap_back == self.multisend_data_swap_back
        assert msg_b is not None and isinstance(msg_b, ContractApiMessage)
        tx_hash_swap_back = cast(str, msg_b.raw_transaction.body["tx_hash"])[2:]
        assert tx_hash_swap_back == self.most_voted_tx_hash_swap_back

        transfers_exit = cast(ContractApiMessage, transfers_msg_exit).state.body["logs"]

        amount_lp_sent = parse_tx_token_balance(
            cast(list, transfers_exit),
            LP_TOKEN_ADDRESS,
            self.safe_contract_address,
            LP_TOKEN_ADDRESS,
        )
        assert (
            amount_lp_sent == 1000
        ), f"Token LP amount sent is not correct: {amount_lp_sent} != 1000"

        amount_a_received = parse_tx_token_balance(
            cast(list, transfers_exit),
            TOKEN_A_ADDRESS,
            LP_TOKEN_ADDRESS,
            self.safe_contract_address,
        )
        assert (
            amount_a_received == 1000
        ), f"Token A amount received is not correct: {amount_a_received} != 1000"

        amount_b_received = parse_tx_token_balance(
            cast(list, transfers_exit),
            TOKEN_B_ADDRESS,
            LP_TOKEN_ADDRESS,
            self.safe_contract_address,
        )
        assert (
            amount_b_received == 1000
        ), f"Token B amount received is not correct: {amount_b_received} != 1000"

        payload = hash_payload_to_hex(
            tx_hash_swap_back,
            0,
            SAFE_TX_GAS_SWAP_BACK,
            self.multisend_contract_address,
            bytes.fromhex(self.multisend_data_swap_back),
            SafeOperation.DELEGATE_CALL.value,
        )

        # update period state with safe's tx hash
        self.tx_settlement_synchronized_data.update(
            most_voted_tx_hash=payload,
        )

        # Sign, send, validate
        self.send_and_validate()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'abci' skill."""
import binascii
import pprint
from abc import ABC
from collections import deque
from typing import (
    Any,
    Deque,
    Dict,
    Generator,
    Iterator,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from aea.protocols.base import Message
from web3.types import Nonce, TxData, Wei

from packages.valory.contracts.gnosis_safe.contract import GnosisSafeContract
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.skills.abstract_round_abci.behaviour_utils import RPCResponseStatus
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.common import (
    RandomnessBehaviour,
    SelectKeeperBehaviour,
)
from packages.valory.skills.abstract_round_abci.utils import VerifyDrand
from packages.valory.skills.transaction_settlement_abci.models import TransactionParams
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
    skill_input_hex_to_payload,
    tx_hist_payload_to_hex,
)
from packages.valory.skills.transaction_settlement_abci.payloads import (
    CheckTransactionHistoryPayload,
    FinalizationTxPayload,
    RandomnessPayload,
    ResetPayload,
    SelectKeeperPayload,
    SignaturePayload,
    SynchronizeLateMessagesPayload,
    ValidatePayload,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    CheckLateTxHashesRound,
    CheckTransactionHistoryRound,
    CollectSignatureRound,
    FinalizationRound,
    RandomnessTransactionSubmissionRound,
    ResetRound,
    SelectKeeperTransactionSubmissionRoundA,
    SelectKeeperTransactionSubmissionRoundB,
    SelectKeeperTransactionSubmissionRoundBAfterTimeout,
    SynchronizeLateMessagesRound,
    SynchronizedData,
    TransactionSubmissionAbciApp,
    ValidateTransactionRound,
)


TxDataType = Dict[str, Union[VerificationStatus, Deque[str], int, Set[str], str]]


drand_check = VerifyDrand()


class TransactionSettlementBaseBehaviour(BaseBehaviour, ABC):
    """Base behaviour for the common apps' skill."""

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, super().synchronized_data)

    @property
    def params(self) -> TransactionParams:
        """Return the params."""
        return cast(TransactionParams, super().params)

    @staticmethod
    def serialized_keepers(keepers: Deque[str], keeper_retries: int) -> str:
        """Get the keepers serialized."""
        if len(keepers) == 0:
            return ""
        keepers_ = "".join(keepers)
        keeper_retries_ = keeper_retries.to_bytes(32, "big").hex()
        concatenated = keeper_retries_ + keepers_

        return concatenated

    def _get_tx_data(
        self, message: ContractApiMessage
    ) -> Generator[None, None, TxDataType]:
        """Get the transaction data from a `ContractApiMessage`."""
        tx_data: TxDataType = {
            "status": VerificationStatus.PENDING,
            "keepers": self.synchronized_data.keepers,
            "keeper_retries": self.synchronized_data.keeper_retries,
            "blacklisted_keepers": self.synchronized_data.blacklisted_keepers,
            "tx_digest": "",
        }

        if (
            message.performative == ContractApiMessage.Performative.ERROR
            and message.message is not None
        ):
            if self._safe_nonce_reused(message.message):
                tx_data["status"] = VerificationStatus.VERIFIED
            else:
                tx_data["status"] = VerificationStatus.ERROR
            self.context.logger.warning(
                f"get_raw_safe_transaction unsuccessful! Received: {message}"
            )
            return tx_data

        if message.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
            self.context.logger.warning(
                f"get_raw_safe_transaction unsuccessful! Received: {message}"
            )
            return tx_data

        tx_digest, rpc_status = yield from self.send_raw_transaction(
            message.raw_transaction
        )

        if rpc_status == RPCResponseStatus.INCORRECT_NONCE:
            tx_data["status"] = VerificationStatus.ERROR
            self.context.logger.warning(
                "send_raw_transaction unsuccessful! Incorrect nonce."
            )

        if rpc_status == RPCResponseStatus.INSUFFICIENT_FUNDS:
            # blacklist self.
            tx_data["status"] = VerificationStatus.INSUFFICIENT_FUNDS
            blacklisted = cast(Deque[str], tx_data["keepers"]).popleft()
            tx_data["keeper_retries"] = 1
            cast(Set[str], tx_data["blacklisted_keepers"]).add(blacklisted)
            self.context.logger.warning(
                "send_raw_transaction unsuccessful! Insufficient funds."
            )

        if rpc_status != RPCResponseStatus.SUCCESS:
            self.context.logger.warning(
                f"send_raw_transaction unsuccessful! Received: {rpc_status}"
            )
            return tx_data

        tx_data["tx_digest"] = cast(str, tx_digest)

        nonce = Nonce(int(cast(str, message.raw_transaction.body["nonce"])))
        gas_price = {
            gas_price_param: Wei(
                int(
                    cast(
                        str,
                        message.raw_transaction.body[gas_price_param],
                    )
                )
            )
            for gas_price_param in ("maxPriorityFeePerGas", "maxFeePerGas")
        }
        # Set hash, nonce and tip.
        self.params.tx_hash = cast(str, tx_data["tx_digest"])
        if nonce == self.params.nonce:
            self.context.logger.info(
                "Attempting to replace transaction "
                f"with old gas price parameters {self.params.gas_price}, using new gas price parameters {gas_price}"
            )
        else:
            self.context.logger.info(
                f"Sent transaction for mining with gas parameters {gas_price}"
            )
            self.params.nonce = nonce
        self.params.gas_price = gas_price

        return tx_data

    def _verify_tx(self, tx_hash: str) -> Generator[None, None, ContractApiMessage]:
        """Verify a transaction."""
        tx_params = skill_input_hex_to_payload(
            self.synchronized_data.most_voted_tx_hash
        )

        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="verify_tx",
            tx_hash=tx_hash,
            owners=tuple(self.synchronized_data.participants),
            to_address=tx_params["to_address"],
            value=tx_params["ether_value"],
            data=tx_params["data"],
            safe_tx_gas=tx_params["safe_tx_gas"],
            signatures_by_owner={
                key: payload.signature
                for key, payload in self.synchronized_data.participant_to_signature.items()
            },
            operation=tx_params["operation"],
        )

        return contract_api_msg

    @staticmethod
    def _safe_nonce_reused(revert_reason: str) -> bool:
        """Check for GS026."""
        return "GS026" in revert_reason


class RandomnessTransactionSubmissionBehaviour(RandomnessBehaviour):
    """Retrieve randomness."""

    behaviour_id = "randomness_transaction_submission"
    matching_round = RandomnessTransactionSubmissionRound
    payload_class = RandomnessPayload


class SelectKeeperTransactionSubmissionBehaviourA(  # pylint: disable=too-many-ancestors
    SelectKeeperBehaviour, TransactionSettlementBaseBehaviour
):
    """Select the keeper agent."""

    behaviour_id = "select_keeper_transaction_submission_a"
    matching_round = SelectKeeperTransactionSubmissionRoundA
    payload_class = SelectKeeperPayload

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            keepers = deque((self._select_keeper(),))
            payload = self.payload_class(
                self.context.agent_address, self.serialized_keepers(keepers, 1)
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class SelectKeeperTransactionSubmissionBehaviourB(  # pylint: disable=too-many-ancestors
    SelectKeeperTransactionSubmissionBehaviourA
):
    """Select the keeper b agent."""

    behaviour_id = "select_keeper_transaction_submission_b"
    matching_round = SelectKeeperTransactionSubmissionRoundB

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
            - If we have not selected enough keepers for the period,
                select a keeper randomly and add it to the keepers' queue, with top priority.
            - Otherwise, cycle through the keepers' subset, using the following logic:
                A `PENDING` verification status means that we have not received any errors,
                therefore, all we know is that the tx has not been mined yet due to low pricing.
                Consequently, we are going to retry with the same keeper in order to replace the transaction.
                However, if we receive a status other than `PENDING`, we need to cycle through the keepers' subset.
                Moreover, if the current keeper has reached the allowed number of retries, then we cycle anyway.
            - Send the transaction with the keepers and wait for it to be mined.
            - Wait until ABCI application transitions to the next round.
            - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            keepers = self.synchronized_data.keepers
            keeper_retries = 1

            if (
                self.synchronized_data.keepers_threshold_exceeded
            ):  # TODO: I think this should be second prio
                keepers.rotate(-1)
                self.context.logger.info(f"Rotated keepers to: {keepers}.")
            elif (
                self.synchronized_data.keeper_retries
                != self.params.keeper_allowed_retries
                and self.synchronized_data.final_verification_status
                == VerificationStatus.PENDING
            ):
                keeper_retries += self.synchronized_data.keeper_retries
                self.context.logger.info(
                    f"Kept keepers and incremented retries: {keepers}."
                )
            else:
                keepers.appendleft(self._select_keeper())

            payload = self.payload_class(
                self.context.agent_address,
                self.serialized_keepers(keepers, keeper_retries),
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class SelectKeeperTransactionSubmissionBehaviourBAfterTimeout(  # pylint: disable=too-many-ancestors
    SelectKeeperTransactionSubmissionBehaviourB
):
    """Select the keeper b agent after a timeout."""

    behaviour_id = "select_keeper_transaction_submission_b_after_timeout"
    matching_round = SelectKeeperTransactionSubmissionRoundBAfterTimeout


class ValidateTransactionBehaviour(TransactionSettlementBaseBehaviour):
    """Validate a transaction."""

    behaviour_id = "validate_transaction"
    matching_round = ValidateTransactionRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Validate that the transaction hash provided by the keeper points to a
          valid transaction.
        - Send the transaction with the validation result and wait for it to be
          mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            is_correct = yield from self.has_transaction_been_sent()
            if is_correct:
                self.context.logger.info(
                    f"Finalized with transaction hash: {self.synchronized_data.to_be_validated_tx_hash}"
                )
            payload = ValidatePayload(self.context.agent_address, is_correct)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def has_transaction_been_sent(self) -> Generator[None, None, Optional[bool]]:
        """Transaction verification."""
        response = yield from self.get_transaction_receipt(
            self.synchronized_data.to_be_validated_tx_hash,
            self.params.retry_timeout,
            self.params.retry_attempts,
        )
        if response is None:  # pragma: nocover
            self.context.logger.error(
                f"tx {self.synchronized_data.to_be_validated_tx_hash} receipt check timed out!"
            )
            return None

        contract_api_msg = yield from self._verify_tx(
            self.synchronized_data.to_be_validated_tx_hash
        )
        if (
            contract_api_msg.performative != ContractApiMessage.Performative.STATE
        ):  # pragma: nocover
            self.context.logger.error(
                f"verify_tx unsuccessful! Received: {contract_api_msg}"
            )
            return False
        verified = cast(bool, contract_api_msg.state.body["verified"])
        verified_log = (
            f"Verified result: {verified}"
            if verified
            else f"Verified result: {verified}, all: {contract_api_msg.state.body}"
        )
        self.context.logger.info(verified_log)
        return verified


CHECK_TX_HISTORY = "check_transaction_history"


class CheckTransactionHistoryBehaviour(TransactionSettlementBaseBehaviour):
    """Check the transaction history."""

    behaviour_id = CHECK_TX_HISTORY
    matching_round = CheckTransactionHistoryRound

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            verification_status, tx_hash = yield from self._check_tx_history()

            if verification_status == VerificationStatus.VERIFIED:
                msg = f"A previous transaction {tx_hash} has already been verified "
                msg += (
                    f"for {self.synchronized_data.to_be_validated_tx_hash}."
                    if self.synchronized_data.tx_hashes_history
                    else "and was synced after the finalization round timed out."
                )
                self.context.logger.info(msg)
            elif verification_status == VerificationStatus.NOT_VERIFIED:
                self.context.logger.info(
                    f"No previous transaction has been verified for "
                    f"{self.synchronized_data.to_be_validated_tx_hash}."
                )

            verified_res = tx_hist_payload_to_hex(verification_status, tx_hash)
            payload = CheckTransactionHistoryPayload(
                self.context.agent_address, verified_res
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def _check_tx_history(
        self,
    ) -> Generator[None, None, Tuple[VerificationStatus, Optional[str]]]:
        """Check the transaction history."""
        history = (
            self.synchronized_data.tx_hashes_history
            if self.behaviour_id == CHECK_TX_HISTORY
            else self.synchronized_data.late_arriving_tx_hashes
        )

        if not history:
            self.context.logger.error(
                "An unexpected error occurred! The data history does not contain any transaction hashes, "
                f"but entered the `{self.behaviour_id}` behaviour."
            )
            return VerificationStatus.ERROR, None

        for tx_hash in history[::-1]:
            contract_api_msg = yield from self._verify_tx(tx_hash)

            if (
                contract_api_msg.performative != ContractApiMessage.Performative.STATE
            ):  # pragma: nocover
                self.context.logger.error(
                    f"verify_tx unsuccessful for {tx_hash}! Received: {contract_api_msg}"
                )
                return VerificationStatus.ERROR, tx_hash

            verified = cast(bool, contract_api_msg.state.body["verified"])
            verified_log = f"Verified result for {tx_hash}: {verified}"

            if verified:
                self.context.logger.info(verified_log)
                return VerificationStatus.VERIFIED, tx_hash

            self.context.logger.info(
                verified_log + f", all: {contract_api_msg.state.body}"
            )

            status = cast(int, contract_api_msg.state.body["status"])
            if status == -1:
                self.context.logger.info(f"Tx hash {tx_hash} has no receipt!")
                continue

            tx_data = cast(TxData, contract_api_msg.state.body["transaction"])
            revert_reason = yield from self._get_revert_reason(tx_data)

            if revert_reason is not None:
                if self._safe_nonce_reused(revert_reason):
                    check_expected_to_be_verified = (
                        "The next tx check"
                        if self.behaviour_id == CHECK_TX_HISTORY
                        else "One of the next tx checks"
                    )
                    self.context.logger.info(
                        f"The safe's nonce has been reused for {tx_hash}. "
                        f"{check_expected_to_be_verified} is expected to be verified!"
                    )
                    continue

                self.context.logger.warning(
                    f"Payload is invalid for {tx_hash}! Cannot continue. Received: {revert_reason}"
                )

            return VerificationStatus.INVALID_PAYLOAD, tx_hash

        return VerificationStatus.NOT_VERIFIED, None

    def _get_revert_reason(self, tx: TxData) -> Generator[None, None, Optional[str]]:
        """Get the revert reason of the given transaction."""
        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="revert_reason",
            tx=tx,
        )

        if (
            contract_api_msg.performative != ContractApiMessage.Performative.STATE
        ):  # pragma: nocover
            self.context.logger.error(
                f"An unexpected error occurred while checking {tx['hash'].hex()}: {contract_api_msg}"
            )
            return None

        return cast(str, contract_api_msg.state.body["revert_reason"])


class CheckLateTxHashesBehaviour(  # pylint: disable=too-many-ancestors
    CheckTransactionHistoryBehaviour
):
    """Check the late-arriving transaction hashes."""

    behaviour_id = "check_late_tx_hashes"
    matching_round = CheckLateTxHashesRound


class SynchronizeLateMessagesBehaviour(TransactionSettlementBaseBehaviour):
    """Synchronize late-arriving messages behaviour."""

    behaviour_id = "sync_late_messages"
    matching_round = SynchronizeLateMessagesRound

    def __init__(self, **kwargs: Any):
        """Initialize a `SynchronizeLateMessagesBehaviour`"""
        super().__init__(**kwargs)
        # if we timed out during finalization, but we managed to receive a tx hash,
        # then we sync it here by initializing the `_tx_hashes` with the unsynced hash.
        self._tx_hashes: str = self.params.tx_hash
        self._messages_iterator: Iterator[ContractApiMessage] = iter(
            self.params.late_messages
        )

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            current_message = next(self._messages_iterator, None)
            if current_message is not None:
                tx_data = yield from self._get_tx_data(current_message)
                self.context.logger.info(
                    f"Found a late arriving message {current_message}. Result data: {tx_data}"
                )
                # here, we concatenate the tx_hashes of all the late-arriving messages. Later, we will parse them.
                self._tx_hashes += cast(str, tx_data["tx_digest"])
                return

            payload = SynchronizeLateMessagesPayload(
                self.context.agent_address, self._tx_hashes
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            # reset the local parameters if we were able to send them.
            self.params.tx_hash = ""
            self.params.late_messages = []
            yield from self.wait_until_round_end()

        self.set_done()


class SignatureBehaviour(TransactionSettlementBaseBehaviour):
    """Signature behaviour."""

    behaviour_id = "sign"
    matching_round = CollectSignatureRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Request the signature of the transaction hash.
        - Send the signature as a transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            self.context.logger.info(
                f"Consensus reached on tx hash: {self.synchronized_data.most_voted_tx_hash}"
            )
            signature_hex = yield from self._get_safe_tx_signature()
            payload = SignaturePayload(self.context.agent_address, signature_hex)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def _get_safe_tx_signature(self) -> Generator[None, None, str]:
        """Get signature of safe transaction hash."""
        tx_params = skill_input_hex_to_payload(
            self.synchronized_data.most_voted_tx_hash
        )
        # is_deprecated_mode=True because we want to call Account.signHash,
        # which is the same used by gnosis-py
        safe_tx_hash_bytes = binascii.unhexlify(tx_params["safe_tx_hash"])
        signature_hex = yield from self.get_signature(
            safe_tx_hash_bytes, is_deprecated_mode=True
        )
        # remove the leading '0x'
        signature_hex = signature_hex[2:]
        self.context.logger.info(f"Signature: {signature_hex}")
        return signature_hex


class FinalizeBehaviour(TransactionSettlementBaseBehaviour):
    """Finalize behaviour."""

    behaviour_id = "finalize"
    matching_round = FinalizationRound

    def _i_am_not_sending(self) -> bool:
        """Indicates if the current agent is the sender or not."""
        return (
            self.context.agent_address
            != self.synchronized_data.most_voted_keeper_address
        )

    def async_act(self) -> Generator[None, None, None]:
        """
        Do the action.

        Steps:
        - If the agent is the keeper, then prepare the transaction and send it.
        - Otherwise, wait until the next round.
        - If a timeout is hit, set exit A event, otherwise set done event.
        """
        if self._i_am_not_sending():
            yield from self._not_sender_act()
        else:
            yield from self._sender_act()

    def _not_sender_act(self) -> Generator:
        """Do the non-sender action."""
        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            self.context.logger.info(
                f"Waiting for the keeper to do its keeping: {self.synchronized_data.most_voted_keeper_address}"
            )
            yield from self.wait_until_round_end()
        self.set_done()

    def _sender_act(self) -> Generator[None, None, None]:
        """Do the sender action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            self.context.logger.info(
                "I am the designated sender, attempting to send the safe transaction..."
            )
            tx_data = yield from self._send_safe_transaction()
            if (
                tx_data["tx_digest"] == ""
                and tx_data["status"] == VerificationStatus.PENDING
            ) or tx_data["status"] == VerificationStatus.ERROR:
                self.context.logger.error(
                    "Did not succeed with finalising the transaction!"
                )
            elif tx_data["status"] == VerificationStatus.VERIFIED:
                self.context.logger.error(
                    "Trying to finalize a transaction which has been verified already!"
                )
            else:
                self.context.logger.info(
                    f"Finalization tx digest: {cast(str, tx_data['tx_digest'])}"
                )
                self.context.logger.debug(
                    f"Signatures: {pprint.pformat(self.synchronized_data.participant_to_signature)}"
                )

            tx_hashes_history = self.synchronized_data.tx_hashes_history
            if tx_data["tx_digest"] != "":
                tx_hashes_history.append(cast(str, tx_data["tx_digest"]))

            tx_data_serialized = {
                "status_value": cast(VerificationStatus, tx_data["status"]).value,
                "serialized_keepers": self.serialized_keepers(
                    cast(Deque[str], tx_data["keepers"]),
                    cast(int, tx_data["keeper_retries"]),
                ),
                "blacklisted_keepers": "".join(
                    cast(Set[str], tx_data["blacklisted_keepers"])
                ),
                "tx_hashes_history": "".join(tx_hashes_history),
                "received_hash": bool(tx_data["tx_digest"]),
            }

            payload = FinalizationTxPayload(
                self.context.agent_address,
                cast(Dict[str, Union[str, int, bool]], tx_data_serialized),
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            # reset the local tx hash parameter if we were able to send it
            self.params.tx_hash = ""
            yield from self.wait_until_round_end()

        self.set_done()

    def _send_safe_transaction(
        self,
    ) -> Generator[None, None, TxDataType]:
        """Send a Safe transaction using the participants' signatures."""
        tx_params = skill_input_hex_to_payload(
            self.synchronized_data.most_voted_tx_hash
        )

        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_raw_safe_transaction",
            sender_address=self.context.agent_address,
            owners=tuple(self.synchronized_data.participants),
            to_address=tx_params["to_address"],
            value=tx_params["ether_value"],
            data=tx_params["data"],
            safe_tx_gas=tx_params["safe_tx_gas"],
            signatures_by_owner={
                key: payload.signature
                for key, payload in self.synchronized_data.participant_to_signature.items()
            },
            nonce=self.params.nonce,
            old_price=self.params.gas_price,
            operation=tx_params["operation"],
        )

        tx_data = yield from self._get_tx_data(contract_api_msg)
        return tx_data

    def handle_late_messages(self, behaviour_id: str, message: Message) -> None:
        """Store a potentially late-arriving message locally.

        :param behaviour_id: the id of the behaviour in which the message belongs to.
        :param message: the late arriving message to handle.
        """
        if (
            isinstance(message, ContractApiMessage)
            and behaviour_id == self.behaviour_id
        ):
            self.context.logger.info(f"Late message arrived: {message}")
            self.params.late_messages.append(message)
        else:
            super().handle_late_messages(behaviour_id, message)


class ResetBehaviour(TransactionSettlementBaseBehaviour):
    """Reset behaviour."""

    matching_round = ResetRound
    behaviour_id = "reset"

    def async_act(self) -> Generator:
        """Do the action."""
        self.context.logger.info(
            f"Period {self.synchronized_data.period_count} was not finished. Resetting!"
        )
        payload = ResetPayload(
            self.context.agent_address, self.synchronized_data.period_count
        )
        yield from self.send_a2a_transaction(payload)
        yield from self.wait_until_round_end()
        self.set_done()


class TransactionSettlementRoundBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the basic transaction settlement."""

    initial_behaviour_cls = RandomnessTransactionSubmissionBehaviour
    abci_app_cls = TransactionSubmissionAbciApp  # type: ignore
    behaviours: Set[Type[BaseBehaviour]] = {
        RandomnessTransactionSubmissionBehaviour,  # type: ignore
        SelectKeeperTransactionSubmissionBehaviourA,  # type: ignore
        SelectKeeperTransactionSubmissionBehaviourB,  # type: ignore
        SelectKeeperTransactionSubmissionBehaviourBAfterTimeout,  # type: ignore
        ValidateTransactionBehaviour,  # type: ignore
        CheckTransactionHistoryBehaviour,  # type: ignore
        SignatureBehaviour,  # type: ignore
        FinalizeBehaviour,  # type: ignore
        SynchronizeLateMessagesBehaviour,  # type: ignore
        CheckLateTxHashesBehaviour,  # type: ignore
        ResetBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Custom objects for the transaction settlement ABCI application."""

from typing import Any, Dict, List, Optional

from web3.types import Nonce, Wei

from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.models import ApiSpecs, BaseParams
from packages.valory.skills.abstract_round_abci.models import (
    BenchmarkTool as BaseBenchmarkTool,
)
from packages.valory.skills.abstract_round_abci.models import Requests as BaseRequests
from packages.valory.skills.abstract_round_abci.models import (
    SharedState as BaseSharedState,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    TransactionSubmissionAbciApp,
)


BenchmarkTool = BaseBenchmarkTool


class SharedState(BaseSharedState):
    """Keep the current shared state of the skill."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the state."""
        super().__init__(*args, abci_app_cls=TransactionSubmissionAbciApp, **kwargs)


class TransactionParams(BaseParams):
    """Transaction settlement agent-specific parameters."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the parameters object.

        We keep track of the nonce and tip across rounds and periods.
        We reuse it each time a new raw transaction is generated. If
        at the time of the new raw transaction being generated the nonce
        on the ledger does not match the nonce on the skill, then we ignore
        the skill nonce and tip (effectively we price fresh). Otherwise, we
        are in a re-submission scenario where we need to take account of the
        old tip.

        :param args: positional arguments
        :param kwargs: keyword arguments
        """
        self.tx_hash: str = ""
        self.nonce: Optional[Nonce] = None
        self.gas_price: Optional[Dict[str, Wei]] = None
        self.late_messages: List[ContractApiMessage] = []
        self.keeper_allowed_retries: int = self._ensure(
            "keeper_allowed_retries", kwargs
        )
        self.validate_timeout = self._ensure("validate_timeout", kwargs)
        self.finalize_timeout = self._ensure("finalize_timeout", kwargs)
        self.history_check_timeout = self._ensure("history_check_timeout", kwargs)
        super().__init__(*args, **kwargs)


class RandomnessApi(ApiSpecs):
    """A model that wraps ApiSpecs for randomness api specifications."""


Requests = BaseRequests



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains abci's message definition."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,too-many-branches,not-an-iterable,unidiomatic-typecheck,unsubscriptable-object
import logging
from typing import Any, Optional, Set, Tuple, cast

from aea.configurations.base import PublicId
from aea.exceptions import AEAEnforceError, enforce
from aea.protocols.base import Message

from packages.valory.protocols.abci.custom_types import CheckTxType as CustomCheckTxType
from packages.valory.protocols.abci.custom_types import (
    ConsensusParams as CustomConsensusParams,
)
from packages.valory.protocols.abci.custom_types import Events as CustomEvents
from packages.valory.protocols.abci.custom_types import Evidences as CustomEvidences
from packages.valory.protocols.abci.custom_types import Header as CustomHeader
from packages.valory.protocols.abci.custom_types import (
    LastCommitInfo as CustomLastCommitInfo,
)
from packages.valory.protocols.abci.custom_types import ProofOps as CustomProofOps
from packages.valory.protocols.abci.custom_types import Result as CustomResult
from packages.valory.protocols.abci.custom_types import SnapShots as CustomSnapShots
from packages.valory.protocols.abci.custom_types import Snapshot as CustomSnapshot
from packages.valory.protocols.abci.custom_types import Timestamp as CustomTimestamp
from packages.valory.protocols.abci.custom_types import (
    ValidatorUpdates as CustomValidatorUpdates,
)


_default_logger = logging.getLogger("aea.packages.valory.protocols.abci.message")

DEFAULT_BODY_SIZE = 4


class AbciMessage(Message):
    """A protocol for ABCI requests and responses."""

    protocol_id = PublicId.from_str("valory/abci:0.1.0")
    protocol_specification_id = PublicId.from_str("valory/abci:0.1.0")

    CheckTxType = CustomCheckTxType

    ConsensusParams = CustomConsensusParams

    Events = CustomEvents

    Evidences = CustomEvidences

    Header = CustomHeader

    LastCommitInfo = CustomLastCommitInfo

    ProofOps = CustomProofOps

    Result = CustomResult

    SnapShots = CustomSnapShots

    Snapshot = CustomSnapshot

    Timestamp = CustomTimestamp

    ValidatorUpdates = CustomValidatorUpdates

    class Performative(Message.Performative):
        """Performatives for the abci protocol."""

        DUMMY = "dummy"
        REQUEST_APPLY_SNAPSHOT_CHUNK = "request_apply_snapshot_chunk"
        REQUEST_BEGIN_BLOCK = "request_begin_block"
        REQUEST_CHECK_TX = "request_check_tx"
        REQUEST_COMMIT = "request_commit"
        REQUEST_DELIVER_TX = "request_deliver_tx"
        REQUEST_ECHO = "request_echo"
        REQUEST_END_BLOCK = "request_end_block"
        REQUEST_FLUSH = "request_flush"
        REQUEST_INFO = "request_info"
        REQUEST_INIT_CHAIN = "request_init_chain"
        REQUEST_LIST_SNAPSHOTS = "request_list_snapshots"
        REQUEST_LOAD_SNAPSHOT_CHUNK = "request_load_snapshot_chunk"
        REQUEST_OFFER_SNAPSHOT = "request_offer_snapshot"
        REQUEST_QUERY = "request_query"
        REQUEST_SET_OPTION = "request_set_option"
        RESPONSE_APPLY_SNAPSHOT_CHUNK = "response_apply_snapshot_chunk"
        RESPONSE_BEGIN_BLOCK = "response_begin_block"
        RESPONSE_CHECK_TX = "response_check_tx"
        RESPONSE_COMMIT = "response_commit"
        RESPONSE_DELIVER_TX = "response_deliver_tx"
        RESPONSE_ECHO = "response_echo"
        RESPONSE_END_BLOCK = "response_end_block"
        RESPONSE_EXCEPTION = "response_exception"
        RESPONSE_FLUSH = "response_flush"
        RESPONSE_INFO = "response_info"
        RESPONSE_INIT_CHAIN = "response_init_chain"
        RESPONSE_LIST_SNAPSHOTS = "response_list_snapshots"
        RESPONSE_LOAD_SNAPSHOT_CHUNK = "response_load_snapshot_chunk"
        RESPONSE_OFFER_SNAPSHOT = "response_offer_snapshot"
        RESPONSE_QUERY = "response_query"
        RESPONSE_SET_OPTION = "response_set_option"

        def __str__(self) -> str:
            """Get the string representation."""
            return str(self.value)

    _performatives = {
        "dummy",
        "request_apply_snapshot_chunk",
        "request_begin_block",
        "request_check_tx",
        "request_commit",
        "request_deliver_tx",
        "request_echo",
        "request_end_block",
        "request_flush",
        "request_info",
        "request_init_chain",
        "request_list_snapshots",
        "request_load_snapshot_chunk",
        "request_offer_snapshot",
        "request_query",
        "request_set_option",
        "response_apply_snapshot_chunk",
        "response_begin_block",
        "response_check_tx",
        "response_commit",
        "response_deliver_tx",
        "response_echo",
        "response_end_block",
        "response_exception",
        "response_flush",
        "response_info",
        "response_init_chain",
        "response_list_snapshots",
        "response_load_snapshot_chunk",
        "response_offer_snapshot",
        "response_query",
        "response_set_option",
    }
    __slots__: Tuple[str, ...] = tuple()

    class _SlotsCls:
        __slots__ = (
            "app_hash",
            "app_state_bytes",
            "app_version",
            "block_version",
            "byzantine_validators",
            "chain_id",
            "chunk",
            "chunk_index",
            "chunk_sender",
            "code",
            "codespace",
            "consensus_param_updates",
            "consensus_params",
            "data",
            "dialogue_reference",
            "dummy_consensus_params",
            "error",
            "events",
            "format",
            "gas_used",
            "gas_wanted",
            "hash",
            "header",
            "height",
            "index",
            "info",
            "info_data",
            "initial_height",
            "key",
            "last_block_app_hash",
            "last_block_height",
            "last_commit_info",
            "log",
            "message",
            "message_id",
            "option_key",
            "option_value",
            "p2p_version",
            "path",
            "performative",
            "proof_ops",
            "prove",
            "query_data",
            "refetch_chunks",
            "reject_senders",
            "result",
            "retain_height",
            "snapshot",
            "snapshots",
            "target",
            "time",
            "tx",
            "type",
            "validator_updates",
            "validators",
            "value",
            "version",
        )

    def __init__(
        self,
        performative: Performative,
        dialogue_reference: Tuple[str, str] = ("", ""),
        message_id: int = 1,
        target: int = 0,
        **kwargs: Any,
    ):
        """
        Initialise an instance of AbciMessage.

        :param message_id: the message id.
        :param dialogue_reference: the dialogue reference.
        :param target: the message target.
        :param performative: the message performative.
        :param **kwargs: extra options.
        """
        super().__init__(
            dialogue_reference=dialogue_reference,
            message_id=message_id,
            target=target,
            performative=AbciMessage.Performative(performative),
            **kwargs,
        )

    @property
    def valid_performatives(self) -> Set[str]:
        """Get valid performatives."""
        return self._performatives

    @property
    def dialogue_reference(self) -> Tuple[str, str]:
        """Get the dialogue_reference of the message."""
        enforce(self.is_set("dialogue_reference"), "dialogue_reference is not set.")
        return cast(Tuple[str, str], self.get("dialogue_reference"))

    @property
    def message_id(self) -> int:
        """Get the message_id of the message."""
        enforce(self.is_set("message_id"), "message_id is not set.")
        return cast(int, self.get("message_id"))

    @property
    def performative(self) -> Performative:  # type: ignore # noqa: F821
        """Get the performative of the message."""
        enforce(self.is_set("performative"), "performative is not set.")
        return cast(AbciMessage.Performative, self.get("performative"))

    @property
    def target(self) -> int:
        """Get the target of the message."""
        enforce(self.is_set("target"), "target is not set.")
        return cast(int, self.get("target"))

    @property
    def app_hash(self) -> bytes:
        """Get the 'app_hash' content from the message."""
        enforce(self.is_set("app_hash"), "'app_hash' content is not set.")
        return cast(bytes, self.get("app_hash"))

    @property
    def app_state_bytes(self) -> bytes:
        """Get the 'app_state_bytes' content from the message."""
        enforce(self.is_set("app_state_bytes"), "'app_state_bytes' content is not set.")
        return cast(bytes, self.get("app_state_bytes"))

    @property
    def app_version(self) -> int:
        """Get the 'app_version' content from the message."""
        enforce(self.is_set("app_version"), "'app_version' content is not set.")
        return cast(int, self.get("app_version"))

    @property
    def block_version(self) -> int:
        """Get the 'block_version' content from the message."""
        enforce(self.is_set("block_version"), "'block_version' content is not set.")
        return cast(int, self.get("block_version"))

    @property
    def byzantine_validators(self) -> CustomEvidences:
        """Get the 'byzantine_validators' content from the message."""
        enforce(
            self.is_set("byzantine_validators"),
            "'byzantine_validators' content is not set.",
        )
        return cast(CustomEvidences, self.get("byzantine_validators"))

    @property
    def chain_id(self) -> str:
        """Get the 'chain_id' content from the message."""
        enforce(self.is_set("chain_id"), "'chain_id' content is not set.")
        return cast(str, self.get("chain_id"))

    @property
    def chunk(self) -> bytes:
        """Get the 'chunk' content from the message."""
        enforce(self.is_set("chunk"), "'chunk' content is not set.")
        return cast(bytes, self.get("chunk"))

    @property
    def chunk_index(self) -> int:
        """Get the 'chunk_index' content from the message."""
        enforce(self.is_set("chunk_index"), "'chunk_index' content is not set.")
        return cast(int, self.get("chunk_index"))

    @property
    def chunk_sender(self) -> str:
        """Get the 'chunk_sender' content from the message."""
        enforce(self.is_set("chunk_sender"), "'chunk_sender' content is not set.")
        return cast(str, self.get("chunk_sender"))

    @property
    def code(self) -> int:
        """Get the 'code' content from the message."""
        enforce(self.is_set("code"), "'code' content is not set.")
        return cast(int, self.get("code"))

    @property
    def codespace(self) -> str:
        """Get the 'codespace' content from the message."""
        enforce(self.is_set("codespace"), "'codespace' content is not set.")
        return cast(str, self.get("codespace"))

    @property
    def consensus_param_updates(self) -> Optional[CustomConsensusParams]:
        """Get the 'consensus_param_updates' content from the message."""
        return cast(
            Optional[CustomConsensusParams], self.get("consensus_param_updates")
        )

    @property
    def consensus_params(self) -> Optional[CustomConsensusParams]:
        """Get the 'consensus_params' content from the message."""
        return cast(Optional[CustomConsensusParams], self.get("consensus_params"))

    @property
    def data(self) -> bytes:
        """Get the 'data' content from the message."""
        enforce(self.is_set("data"), "'data' content is not set.")
        return cast(bytes, self.get("data"))

    @property
    def dummy_consensus_params(self) -> CustomConsensusParams:
        """Get the 'dummy_consensus_params' content from the message."""
        enforce(
            self.is_set("dummy_consensus_params"),
            "'dummy_consensus_params' content is not set.",
        )
        return cast(CustomConsensusParams, self.get("dummy_consensus_params"))

    @property
    def error(self) -> str:
        """Get the 'error' content from the message."""
        enforce(self.is_set("error"), "'error' content is not set.")
        return cast(str, self.get("error"))

    @property
    def events(self) -> CustomEvents:
        """Get the 'events' content from the message."""
        enforce(self.is_set("events"), "'events' content is not set.")
        return cast(CustomEvents, self.get("events"))

    @property
    def format(self) -> int:
        """Get the 'format' content from the message."""
        enforce(self.is_set("format"), "'format' content is not set.")
        return cast(int, self.get("format"))

    @property
    def gas_used(self) -> int:
        """Get the 'gas_used' content from the message."""
        enforce(self.is_set("gas_used"), "'gas_used' content is not set.")
        return cast(int, self.get("gas_used"))

    @property
    def gas_wanted(self) -> int:
        """Get the 'gas_wanted' content from the message."""
        enforce(self.is_set("gas_wanted"), "'gas_wanted' content is not set.")
        return cast(int, self.get("gas_wanted"))

    @property
    def hash(self) -> bytes:
        """Get the 'hash' content from the message."""
        enforce(self.is_set("hash"), "'hash' content is not set.")
        return cast(bytes, self.get("hash"))

    @property
    def header(self) -> CustomHeader:
        """Get the 'header' content from the message."""
        enforce(self.is_set("header"), "'header' content is not set.")
        return cast(CustomHeader, self.get("header"))

    @property
    def height(self) -> int:
        """Get the 'height' content from the message."""
        enforce(self.is_set("height"), "'height' content is not set.")
        return cast(int, self.get("height"))

    @property
    def index(self) -> int:
        """Get the 'index' content from the message."""
        enforce(self.is_set("index"), "'index' content is not set.")
        return cast(int, self.get("index"))

    @property
    def info(self) -> str:
        """Get the 'info' content from the message."""
        enforce(self.is_set("info"), "'info' content is not set.")
        return cast(str, self.get("info"))

    @property
    def info_data(self) -> str:
        """Get the 'info_data' content from the message."""
        enforce(self.is_set("info_data"), "'info_data' content is not set.")
        return cast(str, self.get("info_data"))

    @property
    def initial_height(self) -> int:
        """Get the 'initial_height' content from the message."""
        enforce(self.is_set("initial_height"), "'initial_height' content is not set.")
        return cast(int, self.get("initial_height"))

    @property
    def key(self) -> bytes:
        """Get the 'key' content from the message."""
        enforce(self.is_set("key"), "'key' content is not set.")
        return cast(bytes, self.get("key"))

    @property
    def last_block_app_hash(self) -> bytes:
        """Get the 'last_block_app_hash' content from the message."""
        enforce(
            self.is_set("last_block_app_hash"),
            "'last_block_app_hash' content is not set.",
        )
        return cast(bytes, self.get("last_block_app_hash"))

    @property
    def last_block_height(self) -> int:
        """Get the 'last_block_height' content from the message."""
        enforce(
            self.is_set("last_block_height"), "'last_block_height' content is not set."
        )
        return cast(int, self.get("last_block_height"))

    @property
    def last_commit_info(self) -> CustomLastCommitInfo:
        """Get the 'last_commit_info' content from the message."""
        enforce(
            self.is_set("last_commit_info"), "'last_commit_info' content is not set."
        )
        return cast(CustomLastCommitInfo, self.get("last_commit_info"))

    @property
    def log(self) -> str:
        """Get the 'log' content from the message."""
        enforce(self.is_set("log"), "'log' content is not set.")
        return cast(str, self.get("log"))

    @property
    def message(self) -> str:
        """Get the 'message' content from the message."""
        enforce(self.is_set("message"), "'message' content is not set.")
        return cast(str, self.get("message"))

    @property
    def option_key(self) -> str:
        """Get the 'option_key' content from the message."""
        enforce(self.is_set("option_key"), "'option_key' content is not set.")
        return cast(str, self.get("option_key"))

    @property
    def option_value(self) -> str:
        """Get the 'option_value' content from the message."""
        enforce(self.is_set("option_value"), "'option_value' content is not set.")
        return cast(str, self.get("option_value"))

    @property
    def p2p_version(self) -> int:
        """Get the 'p2p_version' content from the message."""
        enforce(self.is_set("p2p_version"), "'p2p_version' content is not set.")
        return cast(int, self.get("p2p_version"))

    @property
    def path(self) -> str:
        """Get the 'path' content from the message."""
        enforce(self.is_set("path"), "'path' content is not set.")
        return cast(str, self.get("path"))

    @property
    def proof_ops(self) -> CustomProofOps:
        """Get the 'proof_ops' content from the message."""
        enforce(self.is_set("proof_ops"), "'proof_ops' content is not set.")
        return cast(CustomProofOps, self.get("proof_ops"))

    @property
    def prove(self) -> bool:
        """Get the 'prove' content from the message."""
        enforce(self.is_set("prove"), "'prove' content is not set.")
        return cast(bool, self.get("prove"))

    @property
    def query_data(self) -> bytes:
        """Get the 'query_data' content from the message."""
        enforce(self.is_set("query_data"), "'query_data' content is not set.")
        return cast(bytes, self.get("query_data"))

    @property
    def refetch_chunks(self) -> Tuple[int, ...]:
        """Get the 'refetch_chunks' content from the message."""
        enforce(self.is_set("refetch_chunks"), "'refetch_chunks' content is not set.")
        return cast(Tuple[int, ...], self.get("refetch_chunks"))

    @property
    def reject_senders(self) -> Tuple[str, ...]:
        """Get the 'reject_senders' content from the message."""
        enforce(self.is_set("reject_senders"), "'reject_senders' content is not set.")
        return cast(Tuple[str, ...], self.get("reject_senders"))

    @property
    def result(self) -> CustomResult:
        """Get the 'result' content from the message."""
        enforce(self.is_set("result"), "'result' content is not set.")
        return cast(CustomResult, self.get("result"))

    @property
    def retain_height(self) -> int:
        """Get the 'retain_height' content from the message."""
        enforce(self.is_set("retain_height"), "'retain_height' content is not set.")
        return cast(int, self.get("retain_height"))

    @property
    def snapshot(self) -> CustomSnapshot:
        """Get the 'snapshot' content from the message."""
        enforce(self.is_set("snapshot"), "'snapshot' content is not set.")
        return cast(CustomSnapshot, self.get("snapshot"))

    @property
    def snapshots(self) -> CustomSnapShots:
        """Get the 'snapshots' content from the message."""
        enforce(self.is_set("snapshots"), "'snapshots' content is not set.")
        return cast(CustomSnapShots, self.get("snapshots"))

    @property
    def time(self) -> CustomTimestamp:
        """Get the 'time' content from the message."""
        enforce(self.is_set("time"), "'time' content is not set.")
        return cast(CustomTimestamp, self.get("time"))

    @property
    def tx(self) -> bytes:
        """Get the 'tx' content from the message."""
        enforce(self.is_set("tx"), "'tx' content is not set.")
        return cast(bytes, self.get("tx"))

    @property
    def type(self) -> CustomCheckTxType:
        """Get the 'type' content from the message."""
        enforce(self.is_set("type"), "'type' content is not set.")
        return cast(CustomCheckTxType, self.get("type"))

    @property
    def validator_updates(self) -> CustomValidatorUpdates:
        """Get the 'validator_updates' content from the message."""
        enforce(
            self.is_set("validator_updates"), "'validator_updates' content is not set."
        )
        return cast(CustomValidatorUpdates, self.get("validator_updates"))

    @property
    def validators(self) -> CustomValidatorUpdates:
        """Get the 'validators' content from the message."""
        enforce(self.is_set("validators"), "'validators' content is not set.")
        return cast(CustomValidatorUpdates, self.get("validators"))

    @property
    def value(self) -> bytes:
        """Get the 'value' content from the message."""
        enforce(self.is_set("value"), "'value' content is not set.")
        return cast(bytes, self.get("value"))

    @property
    def version(self) -> str:
        """Get the 'version' content from the message."""
        enforce(self.is_set("version"), "'version' content is not set.")
        return cast(str, self.get("version"))

    def _is_consistent(self) -> bool:
        """Check that the message follows the abci protocol."""
        try:
            enforce(
                isinstance(self.dialogue_reference, tuple),
                "Invalid type for 'dialogue_reference'. Expected 'tuple'. Found '{}'.".format(
                    type(self.dialogue_reference)
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[0], str),
                "Invalid type for 'dialogue_reference[0]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[0])
                ),
            )
            enforce(
                isinstance(self.dialogue_reference[1], str),
                "Invalid type for 'dialogue_reference[1]'. Expected 'str'. Found '{}'.".format(
                    type(self.dialogue_reference[1])
                ),
            )
            enforce(
                type(self.message_id) is int,
                "Invalid type for 'message_id'. Expected 'int'. Found '{}'.".format(
                    type(self.message_id)
                ),
            )
            enforce(
                type(self.target) is int,
                "Invalid type for 'target'. Expected 'int'. Found '{}'.".format(
                    type(self.target)
                ),
            )

            # Light Protocol Rule 2
            # Check correct performative
            enforce(
                isinstance(self.performative, AbciMessage.Performative),
                "Invalid 'performative'. Expected either of '{}'. Found '{}'.".format(
                    self.valid_performatives, self.performative
                ),
            )

            # Check correct contents
            actual_nb_of_contents = len(self._body) - DEFAULT_BODY_SIZE
            expected_nb_of_contents = 0
            if self.performative == AbciMessage.Performative.REQUEST_ECHO:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.message, str),
                    "Invalid type for content 'message'. Expected 'str'. Found '{}'.".format(
                        type(self.message)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_FLUSH:
                expected_nb_of_contents = 0
            elif self.performative == AbciMessage.Performative.REQUEST_INFO:
                expected_nb_of_contents = 3
                enforce(
                    isinstance(self.version, str),
                    "Invalid type for content 'version'. Expected 'str'. Found '{}'.".format(
                        type(self.version)
                    ),
                )
                enforce(
                    type(self.block_version) is int,
                    "Invalid type for content 'block_version'. Expected 'int'. Found '{}'.".format(
                        type(self.block_version)
                    ),
                )
                enforce(
                    type(self.p2p_version) is int,
                    "Invalid type for content 'p2p_version'. Expected 'int'. Found '{}'.".format(
                        type(self.p2p_version)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_SET_OPTION:
                expected_nb_of_contents = 2
                enforce(
                    isinstance(self.option_key, str),
                    "Invalid type for content 'option_key'. Expected 'str'. Found '{}'.".format(
                        type(self.option_key)
                    ),
                )
                enforce(
                    isinstance(self.option_value, str),
                    "Invalid type for content 'option_value'. Expected 'str'. Found '{}'.".format(
                        type(self.option_value)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_INIT_CHAIN:
                expected_nb_of_contents = 5
                enforce(
                    isinstance(self.time, CustomTimestamp),
                    "Invalid type for content 'time'. Expected 'Timestamp'. Found '{}'.".format(
                        type(self.time)
                    ),
                )
                enforce(
                    isinstance(self.chain_id, str),
                    "Invalid type for content 'chain_id'. Expected 'str'. Found '{}'.".format(
                        type(self.chain_id)
                    ),
                )
                if self.is_set("consensus_params"):
                    expected_nb_of_contents += 1
                    consensus_params = cast(
                        CustomConsensusParams, self.consensus_params
                    )
                    enforce(
                        isinstance(consensus_params, CustomConsensusParams),
                        "Invalid type for content 'consensus_params'. Expected 'ConsensusParams'. Found '{}'.".format(
                            type(consensus_params)
                        ),
                    )
                enforce(
                    isinstance(self.validators, CustomValidatorUpdates),
                    "Invalid type for content 'validators'. Expected 'ValidatorUpdates'. Found '{}'.".format(
                        type(self.validators)
                    ),
                )
                enforce(
                    isinstance(self.app_state_bytes, bytes),
                    "Invalid type for content 'app_state_bytes'. Expected 'bytes'. Found '{}'.".format(
                        type(self.app_state_bytes)
                    ),
                )
                enforce(
                    type(self.initial_height) is int,
                    "Invalid type for content 'initial_height'. Expected 'int'. Found '{}'.".format(
                        type(self.initial_height)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_QUERY:
                expected_nb_of_contents = 4
                enforce(
                    isinstance(self.query_data, bytes),
                    "Invalid type for content 'query_data'. Expected 'bytes'. Found '{}'.".format(
                        type(self.query_data)
                    ),
                )
                enforce(
                    isinstance(self.path, str),
                    "Invalid type for content 'path'. Expected 'str'. Found '{}'.".format(
                        type(self.path)
                    ),
                )
                enforce(
                    type(self.height) is int,
                    "Invalid type for content 'height'. Expected 'int'. Found '{}'.".format(
                        type(self.height)
                    ),
                )
                enforce(
                    isinstance(self.prove, bool),
                    "Invalid type for content 'prove'. Expected 'bool'. Found '{}'.".format(
                        type(self.prove)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_BEGIN_BLOCK:
                expected_nb_of_contents = 4
                enforce(
                    isinstance(self.hash, bytes),
                    "Invalid type for content 'hash'. Expected 'bytes'. Found '{}'.".format(
                        type(self.hash)
                    ),
                )
                enforce(
                    isinstance(self.header, CustomHeader),
                    "Invalid type for content 'header'. Expected 'Header'. Found '{}'.".format(
                        type(self.header)
                    ),
                )
                enforce(
                    isinstance(self.last_commit_info, CustomLastCommitInfo),
                    "Invalid type for content 'last_commit_info'. Expected 'LastCommitInfo'. Found '{}'.".format(
                        type(self.last_commit_info)
                    ),
                )
                enforce(
                    isinstance(self.byzantine_validators, CustomEvidences),
                    "Invalid type for content 'byzantine_validators'. Expected 'Evidences'. Found '{}'.".format(
                        type(self.byzantine_validators)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_CHECK_TX:
                expected_nb_of_contents = 2
                enforce(
                    isinstance(self.tx, bytes),
                    "Invalid type for content 'tx'. Expected 'bytes'. Found '{}'.".format(
                        type(self.tx)
                    ),
                )
                enforce(
                    isinstance(self.type, CustomCheckTxType),
                    "Invalid type for content 'type'. Expected 'CheckTxType'. Found '{}'.".format(
                        type(self.type)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_DELIVER_TX:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.tx, bytes),
                    "Invalid type for content 'tx'. Expected 'bytes'. Found '{}'.".format(
                        type(self.tx)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_END_BLOCK:
                expected_nb_of_contents = 1
                enforce(
                    type(self.height) is int,
                    "Invalid type for content 'height'. Expected 'int'. Found '{}'.".format(
                        type(self.height)
                    ),
                )
            elif self.performative == AbciMessage.Performative.REQUEST_COMMIT:
                expected_nb_of_contents = 0
            elif self.performative == AbciMessage.Performative.REQUEST_LIST_SNAPSHOTS:
                expected_nb_of_contents = 0
            elif self.performative == AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT:
                expected_nb_of_contents = 2
                enforce(
                    isinstance(self.snapshot, CustomSnapshot),
                    "Invalid type for content 'snapshot'. Expected 'Snapshot'. Found '{}'.".format(
                        type(self.snapshot)
                    ),
                )
                enforce(
                    isinstance(self.app_hash, bytes),
                    "Invalid type for content 'app_hash'. Expected 'bytes'. Found '{}'.".format(
                        type(self.app_hash)
                    ),
                )
            elif (
                self.performative
                == AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK
            ):
                expected_nb_of_contents = 3
                enforce(
                    type(self.height) is int,
                    "Invalid type for content 'height'. Expected 'int'. Found '{}'.".format(
                        type(self.height)
                    ),
                )
                enforce(
                    type(self.format) is int,
                    "Invalid type for content 'format'. Expected 'int'. Found '{}'.".format(
                        type(self.format)
                    ),
                )
                enforce(
                    type(self.chunk_index) is int,
                    "Invalid type for content 'chunk_index'. Expected 'int'. Found '{}'.".format(
                        type(self.chunk_index)
                    ),
                )
            elif (
                self.performative
                == AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK
            ):
                expected_nb_of_contents = 3
                enforce(
                    type(self.index) is int,
                    "Invalid type for content 'index'. Expected 'int'. Found '{}'.".format(
                        type(self.index)
                    ),
                )
                enforce(
                    isinstance(self.chunk, bytes),
                    "Invalid type for content 'chunk'. Expected 'bytes'. Found '{}'.".format(
                        type(self.chunk)
                    ),
                )
                enforce(
                    isinstance(self.chunk_sender, str),
                    "Invalid type for content 'chunk_sender'. Expected 'str'. Found '{}'.".format(
                        type(self.chunk_sender)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_EXCEPTION:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.error, str),
                    "Invalid type for content 'error'. Expected 'str'. Found '{}'.".format(
                        type(self.error)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_ECHO:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.message, str),
                    "Invalid type for content 'message'. Expected 'str'. Found '{}'.".format(
                        type(self.message)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_FLUSH:
                expected_nb_of_contents = 0
            elif self.performative == AbciMessage.Performative.RESPONSE_INFO:
                expected_nb_of_contents = 5
                enforce(
                    isinstance(self.info_data, str),
                    "Invalid type for content 'info_data'. Expected 'str'. Found '{}'.".format(
                        type(self.info_data)
                    ),
                )
                enforce(
                    isinstance(self.version, str),
                    "Invalid type for content 'version'. Expected 'str'. Found '{}'.".format(
                        type(self.version)
                    ),
                )
                enforce(
                    type(self.app_version) is int,
                    "Invalid type for content 'app_version'. Expected 'int'. Found '{}'.".format(
                        type(self.app_version)
                    ),
                )
                enforce(
                    type(self.last_block_height) is int,
                    "Invalid type for content 'last_block_height'. Expected 'int'. Found '{}'.".format(
                        type(self.last_block_height)
                    ),
                )
                enforce(
                    isinstance(self.last_block_app_hash, bytes),
                    "Invalid type for content 'last_block_app_hash'. Expected 'bytes'. Found '{}'.".format(
                        type(self.last_block_app_hash)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_SET_OPTION:
                expected_nb_of_contents = 3
                enforce(
                    type(self.code) is int,
                    "Invalid type for content 'code'. Expected 'int'. Found '{}'.".format(
                        type(self.code)
                    ),
                )
                enforce(
                    isinstance(self.log, str),
                    "Invalid type for content 'log'. Expected 'str'. Found '{}'.".format(
                        type(self.log)
                    ),
                )
                enforce(
                    isinstance(self.info, str),
                    "Invalid type for content 'info'. Expected 'str'. Found '{}'.".format(
                        type(self.info)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_INIT_CHAIN:
                expected_nb_of_contents = 2
                if self.is_set("consensus_params"):
                    expected_nb_of_contents += 1
                    consensus_params = cast(
                        CustomConsensusParams, self.consensus_params
                    )
                    enforce(
                        isinstance(consensus_params, CustomConsensusParams),
                        "Invalid type for content 'consensus_params'. Expected 'ConsensusParams'. Found '{}'.".format(
                            type(consensus_params)
                        ),
                    )
                enforce(
                    isinstance(self.validators, CustomValidatorUpdates),
                    "Invalid type for content 'validators'. Expected 'ValidatorUpdates'. Found '{}'.".format(
                        type(self.validators)
                    ),
                )
                enforce(
                    isinstance(self.app_hash, bytes),
                    "Invalid type for content 'app_hash'. Expected 'bytes'. Found '{}'.".format(
                        type(self.app_hash)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_QUERY:
                expected_nb_of_contents = 9
                enforce(
                    type(self.code) is int,
                    "Invalid type for content 'code'. Expected 'int'. Found '{}'.".format(
                        type(self.code)
                    ),
                )
                enforce(
                    isinstance(self.log, str),
                    "Invalid type for content 'log'. Expected 'str'. Found '{}'.".format(
                        type(self.log)
                    ),
                )
                enforce(
                    isinstance(self.info, str),
                    "Invalid type for content 'info'. Expected 'str'. Found '{}'.".format(
                        type(self.info)
                    ),
                )
                enforce(
                    type(self.index) is int,
                    "Invalid type for content 'index'. Expected 'int'. Found '{}'.".format(
                        type(self.index)
                    ),
                )
                enforce(
                    isinstance(self.key, bytes),
                    "Invalid type for content 'key'. Expected 'bytes'. Found '{}'.".format(
                        type(self.key)
                    ),
                )
                enforce(
                    isinstance(self.value, bytes),
                    "Invalid type for content 'value'. Expected 'bytes'. Found '{}'.".format(
                        type(self.value)
                    ),
                )
                enforce(
                    isinstance(self.proof_ops, CustomProofOps),
                    "Invalid type for content 'proof_ops'. Expected 'ProofOps'. Found '{}'.".format(
                        type(self.proof_ops)
                    ),
                )
                enforce(
                    type(self.height) is int,
                    "Invalid type for content 'height'. Expected 'int'. Found '{}'.".format(
                        type(self.height)
                    ),
                )
                enforce(
                    isinstance(self.codespace, str),
                    "Invalid type for content 'codespace'. Expected 'str'. Found '{}'.".format(
                        type(self.codespace)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_BEGIN_BLOCK:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.events, CustomEvents),
                    "Invalid type for content 'events'. Expected 'Events'. Found '{}'.".format(
                        type(self.events)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_CHECK_TX:
                expected_nb_of_contents = 8
                enforce(
                    type(self.code) is int,
                    "Invalid type for content 'code'. Expected 'int'. Found '{}'.".format(
                        type(self.code)
                    ),
                )
                enforce(
                    isinstance(self.data, bytes),
                    "Invalid type for content 'data'. Expected 'bytes'. Found '{}'.".format(
                        type(self.data)
                    ),
                )
                enforce(
                    isinstance(self.log, str),
                    "Invalid type for content 'log'. Expected 'str'. Found '{}'.".format(
                        type(self.log)
                    ),
                )
                enforce(
                    isinstance(self.info, str),
                    "Invalid type for content 'info'. Expected 'str'. Found '{}'.".format(
                        type(self.info)
                    ),
                )
                enforce(
                    type(self.gas_wanted) is int,
                    "Invalid type for content 'gas_wanted'. Expected 'int'. Found '{}'.".format(
                        type(self.gas_wanted)
                    ),
                )
                enforce(
                    type(self.gas_used) is int,
                    "Invalid type for content 'gas_used'. Expected 'int'. Found '{}'.".format(
                        type(self.gas_used)
                    ),
                )
                enforce(
                    isinstance(self.events, CustomEvents),
                    "Invalid type for content 'events'. Expected 'Events'. Found '{}'.".format(
                        type(self.events)
                    ),
                )
                enforce(
                    isinstance(self.codespace, str),
                    "Invalid type for content 'codespace'. Expected 'str'. Found '{}'.".format(
                        type(self.codespace)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_DELIVER_TX:
                expected_nb_of_contents = 8
                enforce(
                    type(self.code) is int,
                    "Invalid type for content 'code'. Expected 'int'. Found '{}'.".format(
                        type(self.code)
                    ),
                )
                enforce(
                    isinstance(self.data, bytes),
                    "Invalid type for content 'data'. Expected 'bytes'. Found '{}'.".format(
                        type(self.data)
                    ),
                )
                enforce(
                    isinstance(self.log, str),
                    "Invalid type for content 'log'. Expected 'str'. Found '{}'.".format(
                        type(self.log)
                    ),
                )
                enforce(
                    isinstance(self.info, str),
                    "Invalid type for content 'info'. Expected 'str'. Found '{}'.".format(
                        type(self.info)
                    ),
                )
                enforce(
                    type(self.gas_wanted) is int,
                    "Invalid type for content 'gas_wanted'. Expected 'int'. Found '{}'.".format(
                        type(self.gas_wanted)
                    ),
                )
                enforce(
                    type(self.gas_used) is int,
                    "Invalid type for content 'gas_used'. Expected 'int'. Found '{}'.".format(
                        type(self.gas_used)
                    ),
                )
                enforce(
                    isinstance(self.events, CustomEvents),
                    "Invalid type for content 'events'. Expected 'Events'. Found '{}'.".format(
                        type(self.events)
                    ),
                )
                enforce(
                    isinstance(self.codespace, str),
                    "Invalid type for content 'codespace'. Expected 'str'. Found '{}'.".format(
                        type(self.codespace)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_END_BLOCK:
                expected_nb_of_contents = 2
                enforce(
                    isinstance(self.validator_updates, CustomValidatorUpdates),
                    "Invalid type for content 'validator_updates'. Expected 'ValidatorUpdates'. Found '{}'.".format(
                        type(self.validator_updates)
                    ),
                )
                if self.is_set("consensus_param_updates"):
                    expected_nb_of_contents += 1
                    consensus_param_updates = cast(
                        CustomConsensusParams, self.consensus_param_updates
                    )
                    enforce(
                        isinstance(consensus_param_updates, CustomConsensusParams),
                        "Invalid type for content 'consensus_param_updates'. Expected 'ConsensusParams'. Found '{}'.".format(
                            type(consensus_param_updates)
                        ),
                    )
                enforce(
                    isinstance(self.events, CustomEvents),
                    "Invalid type for content 'events'. Expected 'Events'. Found '{}'.".format(
                        type(self.events)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_COMMIT:
                expected_nb_of_contents = 2
                enforce(
                    isinstance(self.data, bytes),
                    "Invalid type for content 'data'. Expected 'bytes'. Found '{}'.".format(
                        type(self.data)
                    ),
                )
                enforce(
                    type(self.retain_height) is int,
                    "Invalid type for content 'retain_height'. Expected 'int'. Found '{}'.".format(
                        type(self.retain_height)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.snapshots, CustomSnapShots),
                    "Invalid type for content 'snapshots'. Expected 'SnapShots'. Found '{}'.".format(
                        type(self.snapshots)
                    ),
                )
            elif self.performative == AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.result, CustomResult),
                    "Invalid type for content 'result'. Expected 'Result'. Found '{}'.".format(
                        type(self.result)
                    ),
                )
            elif (
                self.performative
                == AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK
            ):
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.chunk, bytes),
                    "Invalid type for content 'chunk'. Expected 'bytes'. Found '{}'.".format(
                        type(self.chunk)
                    ),
                )
            elif (
                self.performative
                == AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK
            ):
                expected_nb_of_contents = 3
                enforce(
                    isinstance(self.result, CustomResult),
                    "Invalid type for content 'result'. Expected 'Result'. Found '{}'.".format(
                        type(self.result)
                    ),
                )
                enforce(
                    isinstance(self.refetch_chunks, tuple),
                    "Invalid type for content 'refetch_chunks'. Expected 'tuple'. Found '{}'.".format(
                        type(self.refetch_chunks)
                    ),
                )
                enforce(
                    all(type(element) is int for element in self.refetch_chunks),
                    "Invalid type for tuple elements in content 'refetch_chunks'. Expected 'int'.",
                )
                enforce(
                    isinstance(self.reject_senders, tuple),
                    "Invalid type for content 'reject_senders'. Expected 'tuple'. Found '{}'.".format(
                        type(self.reject_senders)
                    ),
                )
                enforce(
                    all(isinstance(element, str) for element in self.reject_senders),
                    "Invalid type for tuple elements in content 'reject_senders'. Expected 'str'.",
                )
            elif self.performative == AbciMessage.Performative.DUMMY:
                expected_nb_of_contents = 1
                enforce(
                    isinstance(self.dummy_consensus_params, CustomConsensusParams),
                    "Invalid type for content 'dummy_consensus_params'. Expected 'ConsensusParams'. Found '{}'.".format(
                        type(self.dummy_consensus_params)
                    ),
                )

            # Check correct content count
            enforce(
                expected_nb_of_contents == actual_nb_of_contents,
                "Incorrect number of contents. Expected {}. Found {}".format(
                    expected_nb_of_contents, actual_nb_of_contents
                ),
            )

            # Light Protocol Rule 3
            if self.message_id == 1:
                enforce(
                    self.target == 0,
                    "Invalid 'target'. Expected 0 (because 'message_id' is 1). Found {}.".format(
                        self.target
                    ),
                )
        except (AEAEnforceError, ValueError, KeyError) as e:
            _default_logger.error(str(e))
            return False

        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the transaction payloads for the APY estimation app."""
from abc import ABC
from enum import Enum
from typing import Any, Dict, Optional, Union, cast

from packages.valory.skills.abstract_round_abci.base import BaseTxPayload


class TransactionType(Enum):
    """Enumeration of transaction types."""

    REGISTRATION = "registration"
    RANDOMNESS = "randomness"
    SELECT_KEEPER = "select_keeper"
    RESET = "reset"
    FETCHING = "fetching"
    TRANSFORMATION = "transformation"
    BATCH_PREPARATION = "batch_preparation"
    PREPROCESS = "preprocess"
    OPTIMIZATION = "optimization"
    TRAINING = "training"
    TESTING = "testing"
    UPDATE = "update"
    ESTIMATION = "estimation"

    def __str__(self) -> str:
        """Get the string value of the transaction type."""
        return self.value


class BaseAPYPayload(BaseTxPayload, ABC):
    """Base class for the simple abci demo."""

    def __hash__(self) -> int:
        """Hash the payload."""
        return hash(tuple(sorted(self.data.items())))


class RandomnessPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'randomness'."""

    transaction_type = TransactionType.RANDOMNESS

    def __init__(
        self, sender: str, round_id: int, randomness: str, **kwargs: Any
    ) -> None:
        """Initialize an 'select_keeper' transaction payload.

        :param sender: the sender (Ethereum) address
        :param round_id: the round id
        :param randomness: the randomness
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._round_id = round_id
        self._randomness = randomness

    @property
    def round_id(self) -> int:
        """Get the round id."""
        return self._round_id

    @property
    def randomness(self) -> str:
        """Get the randomness."""
        return self._randomness

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(round_id=self._round_id, randomness=self._randomness)


class FetchingPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'fetching'."""

    transaction_type = TransactionType.FETCHING

    def __init__(
        self,
        sender: str,
        history: Optional[str],
        latest_observation_timestamp: int,
        **kwargs: Any
    ) -> None:
        """Initialize a 'fetching' transaction payload.

        :param sender: the sender (Ethereum) address
        :param history: the fetched history's hash.
        :param latest_observation_timestamp: the latest observation's timestamp.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._history = history
        self._latest_observation_timestamp = latest_observation_timestamp

    @property
    def history(self) -> Optional[str]:
        """Get the history's hash."""
        return self._history

    @property
    def latest_observation_timestamp(self) -> int:
        """Get the latest observation's timestamp."""
        return self._latest_observation_timestamp

    @property
    def data(self) -> Dict[str, Union[None, str, int]]:
        """Get the data."""
        return {
            "history": self.history if self.history is not None else "",
            "latest_observation_timestamp": self._latest_observation_timestamp,
        }


class TransformationPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'transformation'."""

    transaction_type = TransactionType.TRANSFORMATION

    def __init__(
        self,
        sender: str,
        transformed_history_hash: Optional[str],
        latest_observation_hist_hash: Optional[str],
        **kwargs: Any
    ) -> None:
        """Initialize a 'transformation' transaction payload.

        :param sender: the sender (Ethereum) address
        :param transformed_history_hash: the transformation's history hash.
        :param latest_observation_hist_hash: the latest observation's history hash.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._transformed_history_hash = transformed_history_hash
        self._latest_observation_hist_hash = latest_observation_hist_hash

    @property
    def transformed_history_hash(self) -> Optional[str]:
        """Get the transformation's history hash."""
        return self._transformed_history_hash

    @property
    def latest_observation_hist_hash(self) -> Optional[str]:
        """Get the latest observation's history hash."""
        return self._latest_observation_hist_hash

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return {
            "transformed_history_hash": self.transformed_history_hash
            if self.transformed_history_hash is not None
            else "",
            "latest_observation_hist_hash": self.latest_observation_hist_hash
            if self.latest_observation_hist_hash is not None
            else "",
        }


class PreprocessPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'preprocess'."""

    transaction_type = TransactionType.PREPROCESS

    def __init__(  # pylint: disable=too-many-arguments
        self,
        sender: str,
        train_hash: Optional[str] = None,
        test_hash: Optional[str] = None,
        train_test: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        """Initialize a 'preprocess' transaction payload.

        :param sender: the sender (Ethereum) address
        :param train_hash: the train data hash.
        :param test_hash: the test data hash.
        :param train_test: the train-test concatenated hash.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._train_hash = train_hash
        self._test_hash = test_hash
        self._train_test = train_test

        if self._train_test is not None:
            self._train_hash = self._test_hash = None

    @property
    def train_test_hash(self) -> Optional[str]:
        """Get the training and testing hash concatenation."""
        if self._train_test is None and not any(
            hash_ is None for hash_ in (self._train_hash, self._test_hash)
        ):
            return cast(str, self._train_hash) + cast(str, self._test_hash)
        return self._train_test

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return {
            "train_test": self.train_test_hash
            if self.train_test_hash is not None
            else "",
        }


class BatchPreparationPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'batch_preparation'."""

    transaction_type = TransactionType.BATCH_PREPARATION

    def __init__(
        self, sender: str, prepared_batch: Optional[str], **kwargs: Any
    ) -> None:
        """Initialize a 'batch_preparation' transaction payload.

        :param sender: the sender (Ethereum) address
        :param prepared_batch: the transformation's hash.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._prepared_batch = prepared_batch

    @property
    def prepared_batch(self) -> Optional[str]:
        """Get the prepared batch's hash."""
        return self._prepared_batch

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return (
            {"prepared_batch": self.prepared_batch}
            if self.prepared_batch is not None
            else {}
        )


class OptimizationPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'optimization'."""

    transaction_type = TransactionType.OPTIMIZATION

    def __init__(self, sender: str, best_params: Optional[str], **kwargs: Any) -> None:
        """Initialize an 'optimization' transaction payload.

        :param sender: the sender (Ethereum) address
        :param best_params: the best params of the study.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._best_params = best_params

    @property
    def best_params(self) -> Optional[str]:
        """Get the best params of the optimization's study."""
        return self._best_params

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return {"best_params": self.best_params} if self.best_params is not None else {}


class TrainingPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'training'."""

    transaction_type = TransactionType.TRAINING

    def __init__(self, sender: str, models_hash: Optional[str], **kwargs: Any) -> None:
        """Initialize a 'training' transaction payload.

        :param sender: the sender (Ethereum) address
        :param models_hash: the model's hash.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._models_hash = models_hash

    @property
    def models_hash(self) -> Optional[str]:
        """Get the models' hash."""
        return self._models_hash

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return {"models_hash": self.models_hash} if self.models_hash is not None else {}


class TestingPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'testing'."""

    transaction_type = TransactionType.TESTING

    def __init__(self, sender: str, report_hash: Optional[str], **kwargs: Any) -> None:
        """Initialize a 'testing' transaction payload.

        :param sender: the sender (Ethereum) address
        :param report_hash: the test's report hash.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._report_hash = report_hash

    @property
    def report_hash(self) -> Optional[str]:
        """Get the test's report hash."""
        return self._report_hash

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return {"report_hash": self.report_hash} if self.report_hash is not None else {}


class UpdatePayload(BaseAPYPayload):
    """Represent a transaction payload of type 'update'."""

    transaction_type = TransactionType.UPDATE

    def __init__(
        self, sender: str, updated_models_hash: Optional[str], **kwargs: Any
    ) -> None:
        """Initialize an 'update' transaction payload.

        :param sender: the sender (Ethereum) address
        :param updated_models_hash: the updated model's hash.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._updated_model_hash = updated_models_hash

    @property
    def updated_models_hash(self) -> Optional[str]:
        """Get the updated model's hash."""
        return self._updated_model_hash

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return (
            {"updated_models_hash": self.updated_models_hash}
            if self.updated_models_hash is not None
            else {}
        )


class EstimatePayload(BaseAPYPayload):
    """Represent a transaction payload of type 'estimate'."""

    transaction_type = TransactionType.ESTIMATION

    def __init__(
        self, sender: str, estimations_hash: Optional[str], **kwargs: Any
    ) -> None:
        """Initialize an 'estimate' transaction payload.

        :param sender: the sender (Ethereum) address
        :param estimations_hash: the hash of the estimations.
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._estimations_hash = estimations_hash

    @property
    def estimations_hash(self) -> Optional[str]:
        """Get the estimations' hash."""
        return self._estimations_hash

    @property
    def data(self) -> Dict[str, Optional[str]]:
        """Get the data."""
        return (
            {"estimations_hash": self.estimations_hash}
            if self.estimations_hash is not None
            else {}
        )


class ResetPayload(BaseAPYPayload):
    """Represent a transaction payload of type 'reset'."""

    transaction_type = TransactionType.RESET

    def __init__(self, sender: str, period_count: int, **kwargs: Any) -> None:
        """Initialize an 'reset' transaction payload.

        :param sender: the sender (Ethereum) address
        :param period_count: the period count id
        :param kwargs: the keyword arguments
        """
        super().__init__(sender, **kwargs)
        self._period_count = period_count

    @property
    def period_count(self) -> int:
        """Get the period_count."""
        return self._period_count

    @property
    def data(self) -> Dict:
        """Get the data."""
        return dict(period_count=self.period_count)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for tendermint protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import Any, Dict, cast

from aea.mail.base_pb2 import DialogueMessage
from aea.mail.base_pb2 import Message as ProtobufMessage
from aea.protocols.base import Message, Serializer

from packages.valory.protocols.tendermint import tendermint_pb2
from packages.valory.protocols.tendermint.custom_types import ErrorCode
from packages.valory.protocols.tendermint.message import TendermintMessage


class TendermintSerializer(Serializer):
    """Serialization for the 'tendermint' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Tendermint' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(TendermintMessage, msg)
        message_pb = ProtobufMessage()
        dialogue_message_pb = DialogueMessage()
        tendermint_msg = tendermint_pb2.TendermintMessage()

        dialogue_message_pb.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        dialogue_message_pb.dialogue_starter_reference = dialogue_reference[0]
        dialogue_message_pb.dialogue_responder_reference = dialogue_reference[1]
        dialogue_message_pb.target = msg.target

        performative_id = msg.performative
        if performative_id == TendermintMessage.Performative.TENDERMINT_CONFIG_REQUEST:
            performative = tendermint_pb2.TendermintMessage.Tendermint_Config_Request_Performative()  # type: ignore
            query = msg.query
            performative.query = query
            tendermint_msg.tendermint_config_request.CopyFrom(performative)
        elif (
            performative_id == TendermintMessage.Performative.TENDERMINT_CONFIG_RESPONSE
        ):
            performative = tendermint_pb2.TendermintMessage.Tendermint_Config_Response_Performative()  # type: ignore
            info = msg.info
            performative.info.update(info)
            tendermint_msg.tendermint_config_response.CopyFrom(performative)
        elif performative_id == TendermintMessage.Performative.ERROR:
            performative = tendermint_pb2.TendermintMessage.Error_Performative()  # type: ignore
            error_code = msg.error_code
            ErrorCode.encode(performative.error_code, error_code)
            error_msg = msg.error_msg
            performative.error_msg = error_msg
            info = msg.info
            performative.info.update(info)
            tendermint_msg.error.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        dialogue_message_pb.content = tendermint_msg.SerializeToString()

        message_pb.dialogue_message.CopyFrom(dialogue_message_pb)
        message_bytes = message_pb.SerializeToString()
        return message_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Tendermint' message.

        :param obj: the bytes object.
        :return: the 'Tendermint' message.
        """
        message_pb = ProtobufMessage()
        tendermint_pb = tendermint_pb2.TendermintMessage()
        message_pb.ParseFromString(obj)
        message_id = message_pb.dialogue_message.message_id
        dialogue_reference = (
            message_pb.dialogue_message.dialogue_starter_reference,
            message_pb.dialogue_message.dialogue_responder_reference,
        )
        target = message_pb.dialogue_message.target

        tendermint_pb.ParseFromString(message_pb.dialogue_message.content)
        performative = tendermint_pb.WhichOneof("performative")
        performative_id = TendermintMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == TendermintMessage.Performative.TENDERMINT_CONFIG_REQUEST:
            query = tendermint_pb.tendermint_config_request.query
            performative_content["query"] = query
        elif (
            performative_id == TendermintMessage.Performative.TENDERMINT_CONFIG_RESPONSE
        ):
            info = tendermint_pb.tendermint_config_response.info
            info_dict = dict(info)
            performative_content["info"] = info_dict
        elif performative_id == TendermintMessage.Performative.ERROR:
            pb2_error_code = tendermint_pb.error.error_code
            error_code = ErrorCode.decode(pb2_error_code)
            performative_content["error_code"] = error_code
            error_msg = tendermint_pb.error.error_msg
            performative_content["error_msg"] = error_msg
            info = tendermint_pb.error.info
            info_dict = dict(info)
            performative_content["info"] = info_dict
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return TendermintMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10tendermint.proto\x12\x1c\x61\x65\x61.valory.tendermint.v0_1_0"\x8a\x08\n\x11TendermintMessage\x12S\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x42.aea.valory.tendermint.v0_1_0.TendermintMessage.Error_PerformativeH\x00\x12{\n\x19tendermint_config_request\x18\x06 \x01(\x0b\x32V.aea.valory.tendermint.v0_1_0.TendermintMessage.Tendermint_Config_Request_PerformativeH\x00\x12}\n\x1atendermint_config_response\x18\x07 \x01(\x0b\x32W.aea.valory.tendermint.v0_1_0.TendermintMessage.Tendermint_Config_Response_PerformativeH\x00\x1a\x8e\x01\n\tErrorCode\x12[\n\nerror_code\x18\x01 \x01(\x0e\x32G.aea.valory.tendermint.v0_1_0.TendermintMessage.ErrorCode.ErrorCodeEnum"$\n\rErrorCodeEnum\x12\x13\n\x0fINVALID_REQUEST\x10\x00\x1a\x37\n&Tendermint_Config_Request_Performative\x12\r\n\x05query\x18\x01 \x01(\t\x1a\xc7\x01\n\'Tendermint_Config_Response_Performative\x12o\n\x04info\x18\x01 \x03(\x0b\x32\x61.aea.valory.tendermint.v0_1_0.TendermintMessage.Tendermint_Config_Response_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xff\x01\n\x12\x45rror_Performative\x12M\n\nerror_code\x18\x01 \x01(\x0b\x32\x39.aea.valory.tendermint.v0_1_0.TendermintMessage.ErrorCode\x12\x11\n\terror_msg\x18\x02 \x01(\t\x12Z\n\x04info\x18\x03 \x03(\x0b\x32L.aea.valory.tendermint.v0_1_0.TendermintMessage.Error_Performative.InfoEntry\x1a+\n\tInfoEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0e\n\x0cperformativeb\x06proto3'
)


_TENDERMINTMESSAGE = DESCRIPTOR.message_types_by_name["TendermintMessage"]
_TENDERMINTMESSAGE_ERRORCODE = _TENDERMINTMESSAGE.nested_types_by_name["ErrorCode"]
_TENDERMINTMESSAGE_TENDERMINT_CONFIG_REQUEST_PERFORMATIVE = (
    _TENDERMINTMESSAGE.nested_types_by_name["Tendermint_Config_Request_Performative"]
)
_TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE = (
    _TENDERMINTMESSAGE.nested_types_by_name["Tendermint_Config_Response_Performative"]
)
_TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE_INFOENTRY = (
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE.nested_types_by_name[
        "InfoEntry"
    ]
)
_TENDERMINTMESSAGE_ERROR_PERFORMATIVE = _TENDERMINTMESSAGE.nested_types_by_name[
    "Error_Performative"
]
_TENDERMINTMESSAGE_ERROR_PERFORMATIVE_INFOENTRY = (
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE.nested_types_by_name["InfoEntry"]
)
_TENDERMINTMESSAGE_ERRORCODE_ERRORCODEENUM = (
    _TENDERMINTMESSAGE_ERRORCODE.enum_types_by_name["ErrorCodeEnum"]
)
TendermintMessage = _reflection.GeneratedProtocolMessageType(
    "TendermintMessage",
    (_message.Message,),
    {
        "ErrorCode": _reflection.GeneratedProtocolMessageType(
            "ErrorCode",
            (_message.Message,),
            {
                "DESCRIPTOR": _TENDERMINTMESSAGE_ERRORCODE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.ErrorCode)
            },
        ),
        "Tendermint_Config_Request_Performative": _reflection.GeneratedProtocolMessageType(
            "Tendermint_Config_Request_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _TENDERMINTMESSAGE_TENDERMINT_CONFIG_REQUEST_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Tendermint_Config_Request_Performative)
            },
        ),
        "Tendermint_Config_Response_Performative": _reflection.GeneratedProtocolMessageType(
            "Tendermint_Config_Response_Performative",
            (_message.Message,),
            {
                "InfoEntry": _reflection.GeneratedProtocolMessageType(
                    "InfoEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE_INFOENTRY,
                        "__module__": "tendermint_pb2"
                        # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Tendermint_Config_Response_Performative.InfoEntry)
                    },
                ),
                "DESCRIPTOR": _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Tendermint_Config_Response_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "InfoEntry": _reflection.GeneratedProtocolMessageType(
                    "InfoEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_INFOENTRY,
                        "__module__": "tendermint_pb2"
                        # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Error_Performative.InfoEntry)
                    },
                ),
                "DESCRIPTOR": _TENDERMINTMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _TENDERMINTMESSAGE,
        "__module__": "tendermint_pb2"
        # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage)
    },
)
_sym_db.RegisterMessage(TendermintMessage)
_sym_db.RegisterMessage(TendermintMessage.ErrorCode)
_sym_db.RegisterMessage(TendermintMessage.Tendermint_Config_Request_Performative)
_sym_db.RegisterMessage(TendermintMessage.Tendermint_Config_Response_Performative)
_sym_db.RegisterMessage(
    TendermintMessage.Tendermint_Config_Response_Performative.InfoEntry
)
_sym_db.RegisterMessage(TendermintMessage.Error_Performative)
_sym_db.RegisterMessage(TendermintMessage.Error_Performative.InfoEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE_INFOENTRY._options = None
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE_INFOENTRY._serialized_options = (
        b"8\001"
    )
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_INFOENTRY._options = None
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_INFOENTRY._serialized_options = b"8\001"
    _TENDERMINTMESSAGE._serialized_start = 51
    _TENDERMINTMESSAGE._serialized_end = 1085
    _TENDERMINTMESSAGE_ERRORCODE._serialized_start = 410
    _TENDERMINTMESSAGE_ERRORCODE._serialized_end = 552
    _TENDERMINTMESSAGE_ERRORCODE_ERRORCODEENUM._serialized_start = 516
    _TENDERMINTMESSAGE_ERRORCODE_ERRORCODEENUM._serialized_end = 552
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_REQUEST_PERFORMATIVE._serialized_start = 554
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_REQUEST_PERFORMATIVE._serialized_end = 609
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE._serialized_start = 612
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE._serialized_end = 811
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE_INFOENTRY._serialized_start = (
        768
    )
    _TENDERMINTMESSAGE_TENDERMINT_CONFIG_RESPONSE_PERFORMATIVE_INFOENTRY._serialized_end = (
        811
    )
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE._serialized_start = 814
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE._serialized_end = 1069
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_INFOENTRY._serialized_start = 768
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_INFOENTRY._serialized_end = 811
# @@protoc_insertion_point(module_scope)
